from tmuxp.plugin import TmuxpPlugin


class PluginAfterWindowFinished(TmuxpPlugin):
    def __init__(self):
        self.message = "[+] This is the Tmuxp Test Plugin"

    def after_window_finished(self, window):
        if window.name == "editor":
            window.rename_window("plugin_test_awf")
        elif window.name == "awf_mw_test":
            window.rename_window("plugin_test_awf_mw")
        elif window.name == "awf_mw_test_2":
            window.rename_window("plugin_test_awf_mw_2")
        elif window.name == "mp_test_owc":
            window.rename_window("mp_test_awf")
        else:
            pass



import os

config_unexpanded = {  # shell_command_before is string in some areas
    "session_name": "sampleconfig",
    "start_directory": "/",
    "windows": [
        {
            "window_name": "editor",
            "start_directory": "~",
            "shell_command_before": "source .venv/bin/activate",
            "panes": [
                {"shell_command": ["vim"]},
                {
                    "shell_command_before": ["rbenv local 2.0.0-p0"],
                    "shell_command": ['cowsay "hey"'],
                },
            ],
            "layout": "main-verticle",
        },
        {
            "shell_command_before": "rbenv local 2.0.0-p0",
            "window_name": "logging",
            "panes": [{"shell_command": ["tail -F /var/log/syslog"]}, {}],
        },
        {
            "window_name": "shufu",
            "panes": [
                {
                    "shell_command_before": ["rbenv local 2.0.0-p0"],
                    "shell_command": ["htop"],
                }
            ],
        },
        {"options": {"automatic-rename": True}, "panes": [{"shell_command": ["htop"]}]},
        {"panes": ["top"]},
    ],
}


def config_expanded():
    return {  # shell_command_before is string in some areas
        "session_name": "sampleconfig",
        "start_directory": "/",
        "windows": [
            {
                "window_name": "editor",
                "start_directory": os.path.expanduser("~"),
                "shell_command_before": {
                    "shell_command": [{"cmd": "source .venv/bin/activate"}]
                },
                "panes": [
                    {"shell_command": [{"cmd": "vim"}]},
                    {
                        "shell_command_before": {
                            "shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]
                        },
                        "shell_command": [{"cmd": 'cowsay "hey"'}],
                    },
                ],
                "layout": "main-verticle",
            },
            {
                "shell_command_before": {
                    "shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]
                },
                "window_name": "logging",
                "panes": [
                    {"shell_command": [{"cmd": "tail -F /var/log/syslog"}]},
                    {"shell_command": []},
                ],
            },
            {
                "window_name": "shufu",
                "panes": [
                    {
                        "shell_command_before": {
                            "shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]
                        },
                        "shell_command": [{"cmd": "htop"}],
                    }
                ],
            },
            {
                "options": {"automatic-rename": True},
                "panes": [{"shell_command": [{"cmd": "htop"}]}],
            },
            {"panes": [{"shell_command": [{"cmd": "top"}]}]},
        ],
    }


def config_after():
    return {  # shell_command_before is string in some areas
        "session_name": "sampleconfig",
        "start_directory": "/",
        "windows": [
            {
                "window_name": "editor",
                "start_directory": os.path.expanduser("~"),
                "shell_command_before": {
                    "shell_command": [{"cmd": "source .venv/bin/activate"}]
                },
                "panes": [
                    {
                        "shell_command": [
                            {"cmd": "source .venv/bin/activate"},
                            {"cmd": "vim"},
                        ]
                    },
                    {
                        "shell_command_before": {
                            "shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]
                        },
                        "shell_command": [
                            {"cmd": "source .venv/bin/activate"},
                            {"cmd": "rbenv local 2.0.0-p0"},
                            {"cmd": 'cowsay "hey"'},
                        ],
                    },
                ],
                "layout": "main-verticle",
            },
            {
                "shell_command_before": {
                    "shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]
                },
                "start_directory": "/",
                "window_name": "logging",
                "panes": [
                    {
                        "shell_command": [
                            {"cmd": "rbenv local 2.0.0-p0"},
                            {"cmd": "tail -F /var/log/syslog"},
                        ]
                    },
                    {"shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]},
                ],
            },
            {
                "start_directory": "/",
                "window_name": "shufu",
                "panes": [
                    {
                        "shell_command_before": {
                            "shell_command": [{"cmd": "rbenv local 2.0.0-p0"}]
                        },
                        "shell_command": [
                            {"cmd": "rbenv local 2.0.0-p0"},
                            {"cmd": "htop"},
                        ],
                    }
                ],
            },
            {
                "start_directory": "/",
                "options": {"automatic-rename": True},
                "panes": [{"shell_command": [{"cmd": "htop"}]}],
            },
            {"start_directory": "/", "panes": [{"shell_command": [{"cmd": "top"}]}]},
        ],
    }



"""CLI utilities for tmuxp.

tmuxp.cli
~~~~~~~~~

"""
import logging
import os
import sys

import click

from libtmux.__about__ import __version__ as libtmux_version
from libtmux.common import has_minimum_version
from libtmux.exc import TmuxCommandNotFound
from tmuxp.cli.ls import command_ls

from .. import exc
from ..__about__ import __version__
from ..log import setup_logger
from .convert import command_convert
from .debug_info import command_debug_info
from .edit import command_edit
from .freeze import command_freeze
from .import_config import command_import
from .load import command_load
from .shell import command_shell
from .utils import tmuxp_echo

logger = logging.getLogger(__name__)


@click.group(context_settings={"obj": {}, "help_option_names": ["-h", "--help"]})
@click.version_option(
    __version__,
    "-V",
    "--version",
    message=f"%(prog)s %(version)s, libtmux {libtmux_version}",
)
@click.option(
    "--log-level",
    default="INFO",
    help="Log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
)
def cli(log_level):
    """Manage tmux sessions.

    Pass the "--help" argument to any command to see detailed help.
    See detailed documentation and examples at:
    http://tmuxp.git-pull.com/"""
    try:
        has_minimum_version()
    except TmuxCommandNotFound:
        tmuxp_echo("tmux not found. tmuxp requires you install tmux first.")
        sys.exit()
    except exc.TmuxpException as e:
        tmuxp_echo(e, err=True)
        sys.exit()
    setup_logger(logger=logger, level=log_level.upper())


def startup(config_dir):
    """
    Initialize CLI.

    Parameters
    ----------
    str : get_config_dir(): Config directory to search
    """

    if not os.path.exists(config_dir):
        os.makedirs(config_dir)


# Register sub-commands here
cli.add_command(command_convert)
cli.add_command(command_edit)
cli.add_command(command_debug_info)
cli.add_command(command_load)
cli.add_command(command_ls)
cli.add_command(command_freeze)
cli.add_command(command_shell)
cli.add_command(command_import)



from tmuxp.plugin import TmuxpPlugin


class PluginOnWindowCreate(TmuxpPlugin):
    def __init__(self):
        self.message = "[+] This is the Tmuxp Test Plugin"

    def on_window_create(self, window):
        if window.name == "editor":
            window.rename_window("plugin_test_owc")
        elif window.name == "owc_mw_test":
            window.rename_window("plugin_test_owc_mw")
        elif window.name == "owc_mw_test_2":
            window.rename_window("plugin_test_owc_mw_2")
        elif window.name == "mp_test":
            window.rename_window("mp_test_owc")
        else:
            pass



from .. import utils as test_utils

tmuxinator_yaml = test_utils.read_config_file("config_tmuxinator/test3.yaml")

tmuxinator_dict = {
    "name": "sample",
    "root": "~/test",
    "socket_name": "foo",
    "tmux_options": "-f ~/.tmux.mac.conf",
    "pre": "sudo /etc/rc.d/mysqld start",
    "pre_window": "rbenv shell 2.0.0-p247",
    "windows": [
        {
            "editor": {
                "pre": [
                    'echo "I get run in each pane, ' 'before each pane command!"',
                    None,
                ],
                "layout": "main-vertical",
                "root": "~/test/editor",
                "panes": ["vim", None, "top"],
            }
        },
        {"shell": ["git pull", "git merge"]},
        {
            "guard": {
                "layout": "tiled",
                "pre": [
                    'echo "I get run in each pane."',
                    'echo "Before each pane command!"',
                ],
                "panes": [None, None, None],
            }
        },
        {"database": "bundle exec rails db"},
        {"server": "bundle exec rails s"},
        {"logs": "tail -f log/development.log"},
        {"console": "bundle exec rails c"},
        {"capistrano": None},
        {"server": "ssh user@example.com"},
    ],
}

expected = {
    "session_name": "sample",
    "socket_name": "foo",
    "start_directory": "~/test",
    "config": "~/.tmux.mac.conf",
    "shell_command": "sudo /etc/rc.d/mysqld start",
    "shell_command_before": ["rbenv shell 2.0.0-p247"],
    "windows": [
        {
            "window_name": "editor",
            "shell_command_before": [
                'echo "I get run in each pane, before each pane command!"',
                None,
            ],
            "layout": "main-vertical",
            "start_directory": "~/test/editor",
            "panes": ["vim", None, "top"],
        },
        {"window_name": "shell", "panes": ["git pull", "git merge"]},
        {
            "window_name": "guard",
            "layout": "tiled",
            "shell_command_before": [
                'echo "I get run in each pane."',
                'echo "Before each pane command!"',
            ],
            "panes": [None, None, None],
        },
        {"window_name": "database", "panes": ["bundle exec rails db"]},
        {"window_name": "server", "panes": ["bundle exec rails s"]},
        {"window_name": "logs", "panes": ["tail -f log/development.log"]},
        {"window_name": "console", "panes": ["bundle exec rails c"]},
        {"window_name": "capistrano", "panes": [None]},
        {"window_name": "server", "panes": ["ssh user@example.com"]},
    ],
}



import os

import click

from .constants import VALID_CONFIG_DIR_FILE_EXTENSIONS
from .utils import get_config_dir


@click.command(
    name="ls",
    short_help="List configured sessions in :meth:`tmuxp.cli.utils.get_config_dir`.",
)
def command_ls():
    tmuxp_dir = get_config_dir()
    if os.path.exists(tmuxp_dir) and os.path.isdir(tmuxp_dir):
        for f in sorted(os.listdir(tmuxp_dir)):
            stem, ext = os.path.splitext(f)
            if os.path.isdir(f) or ext not in VALID_CONFIG_DIR_FILE_EXTENSIONS:
                continue
            print(stem)



from .. import utils as test_utils

teamocil_yaml = test_utils.read_config_file("config_teamocil/test3.yaml")

teamocil_dict = {
    "windows": [
        {
            "name": "my-first-window",
            "root": "~/Projects/foo-www",
            "layout": "even-vertical",
            "filters": {
                "before": "rbenv local 2.0.0-p0",
                "after": "echo 'I am done initializing this pane.'",
            },
            "panes": [
                {"cmd": "git status"},
                {"cmd": "bundle exec rails server --port 40", "focus": True},
                {"cmd": ["sudo service memcached start", "sudo service mongodb start"]},
            ],
        }
    ]
}

expected = {
    "session_name": None,
    "windows": [
        {
            "window_name": "my-first-window",
            "layout": "even-vertical",
            "start_directory": "~/Projects/foo-www",
            "shell_command_before": "rbenv local 2.0.0-p0",
            "shell_command_after": ("echo " "'I am done initializing this pane.'"),
            "panes": [
                {"shell_command": "git status"},
                {"shell_command": "bundle exec rails server --port 40", "focus": True},
                {
                    "shell_command": [
                        "sudo service memcached start",
                        "sudo service mongodb start",
                    ]
                },
            ],
        }
    ],
}



"""Test for tmuxp configuration import, inlining, expanding and export."""
import os
import pathlib
import types
import typing
from typing import Union

import pytest

from tmuxp import config, exc
from tmuxp.config_reader import ConfigReader

from .constants import EXAMPLE_PATH

if typing.TYPE_CHECKING:
    from .fixtures.structures import ConfigTestData


@pytest.fixture
def config_fixture():
    """Deferred import of tmuxp.tests.fixtures.*

    pytest setup (conftest.py) patches os.environ["HOME"], delay execution of
    os.path.expanduser until here.
    """
    from .fixtures import config as test_config_data
    from .fixtures.structures import ConfigTestData

    return ConfigTestData(
        **{
            k: v
            for k, v in test_config_data.__dict__.items()
            if isinstance(v, types.ModuleType)
        }
    )


def load_yaml(path: Union[str, pathlib.Path]) -> str:
    return ConfigReader._from_file(
        pathlib.Path(path) if isinstance(path, str) else path
    )


def load_config(path: Union[str, pathlib.Path]) -> str:
    return ConfigReader._from_file(
        pathlib.Path(path) if isinstance(path, str) else path
    )


def test_export_json(tmp_path: pathlib.Path, config_fixture: "ConfigTestData"):
    json_config_file = tmp_path / "config.json"

    configparser = ConfigReader(config_fixture.sampleconfig.sampleconfigdict)

    json_config_data = configparser.dump("json", indent=2)

    json_config_file.write_text(json_config_data, encoding="utf-8")

    new_config_data = ConfigReader._from_file(path=json_config_file)
    assert config_fixture.sampleconfig.sampleconfigdict == new_config_data


def test_export_yaml(tmp_path: pathlib.Path, config_fixture: "ConfigTestData"):
    yaml_config_file = tmp_path / "config.yaml"

    sampleconfig = config.inline(config_fixture.sampleconfig.sampleconfigdict)
    configparser = ConfigReader(sampleconfig)

    yaml_config_data = configparser.dump("yaml", indent=2, default_flow_style=False)

    yaml_config_file.write_text(yaml_config_data, encoding="utf-8")

    new_config_data = load_config(str(yaml_config_file))
    assert config_fixture.sampleconfig.sampleconfigdict == new_config_data


def test_scan_config(tmp_path: pathlib.Path):
    configs = []

    garbage_file = tmp_path / "config.psd"
    garbage_file.write_text("wat", encoding="utf-8")

    for r, d, f in os.walk(str(tmp_path)):
        for filela in (x for x in f if x.endswith((".json", ".ini", "yaml"))):
            configs.append(str(tmp_path / filela))

    files = 0
    config_json = tmp_path / "config.json"
    config_yaml = tmp_path / "config.yaml"
    config_ini = tmp_path / "config.ini"
    if config_json.exists():
        files += 1
        assert str(config_json) in configs

    if config_yaml.exists():
        files += 1
        assert str(config_yaml) in configs

    if config_ini.exists():
        files += 1
        assert str(config_ini) in configs

    assert len(configs) == files


def test_config_expand1(config_fixture: "ConfigTestData"):
    """Expand shell commands from string to list."""
    test_config = config.expand(config_fixture.expand1.before_config)
    assert test_config == config_fixture.expand1.after_config()


def test_config_expand2(config_fixture: "ConfigTestData"):
    """Expand shell commands from string to list."""
    unexpanded_dict = ConfigReader._load(
        format="yaml", content=config_fixture.expand2.unexpanded_yaml()
    )
    expanded_dict = ConfigReader._load(
        format="yaml", content=config_fixture.expand2.expanded_yaml()
    )
    assert config.expand(unexpanded_dict) == expanded_dict


"""Tests for :meth:`config.inline()`."""

ibefore_config = {  # inline config
    "session_name": "sampleconfig",
    "start_directory": "~",
    "windows": [
        {
            "shell_command": ["top"],
            "window_name": "editor",
            "panes": [{"shell_command": ["vim"]}, {"shell_command": ['cowsay "hey"']}],
            "layout": "main-verticle",
        },
        {
            "window_name": "logging",
            "panes": [{"shell_command": ["tail -F /var/log/syslog"]}],
        },
        {"options": {"automatic-rename": True}, "panes": [{"shell_command": ["htop"]}]},
    ],
}

iafter_config = {
    "session_name": "sampleconfig",
    "start_directory": "~",
    "windows": [
        {
            "shell_command": "top",
            "window_name": "editor",
            "panes": ["vim", 'cowsay "hey"'],
            "layout": "main-verticle",
        },
        {"window_name": "logging", "panes": ["tail -F /var/log/syslog"]},
        {"options": {"automatic-rename": True}, "panes": ["htop"]},
    ],
}


def test_inline_config():
    """:meth:`config.inline()` shell commands list to string."""

    test_config = config.inline(ibefore_config)
    assert test_config == iafter_config


"""Test config inheritance for the nested 'start_command'."""

inheritance_config_before = {
    "session_name": "sampleconfig",
    "start_directory": "/",
    "windows": [
        {
            "window_name": "editor",
            "start_directory": "~",
            "panes": [{"shell_command": ["vim"]}, {"shell_command": ['cowsay "hey"']}],
            "layout": "main-verticle",
        },
        {
            "window_name": "logging",
            "panes": [{"shell_command": ["tail -F /var/log/syslog"]}],
        },
        {"window_name": "shufu", "panes": [{"shell_command": ["htop"]}]},
        {"options": {"automatic-rename": True}, "panes": [{"shell_command": ["htop"]}]},
    ],
}

inheritance_config_after = {
    "session_name": "sampleconfig",
    "start_directory": "/",
    "windows": [
        {
            "window_name": "editor",
            "start_directory": "~",
            "panes": [{"shell_command": ["vim"]}, {"shell_command": ['cowsay "hey"']}],
            "layout": "main-verticle",
        },
        {
            "window_name": "logging",
            "panes": [{"shell_command": ["tail -F /var/log/syslog"]}],
        },
        {"window_name": "shufu", "panes": [{"shell_command": ["htop"]}]},
        {"options": {"automatic-rename": True}, "panes": [{"shell_command": ["htop"]}]},
    ],
}


def test_inheritance_config():
    config = inheritance_config_before

    # TODO: Look at verifying window_start_directory
    # if 'start_directory' in config:
    #     session_start_directory = config['start_directory']
    # else:
    #     session_start_directory = None

    # for windowconfitem in config['windows']:
    #     window_start_directory = None
    #
    #     if 'start_directory' in windowconfitem:
    #         window_start_directory = windowconfitem['start_directory']
    #     elif session_start_directory:
    #         window_start_directory = session_start_directory
    #
    #     for paneconfitem in windowconfitem['panes']:
    #         if 'start_directory' in paneconfitem:
    #             pane_start_directory = paneconfitem['start_directory']
    #         elif window_start_directory:
    #             paneconfitem['start_directory'] = window_start_directory
    #         elif session_start_directory:
    #             paneconfitem['start_directory'] = session_start_directory

    assert config == inheritance_config_after


def test_shell_command_before(config_fixture: "ConfigTestData"):
    """Config inheritance for the nested 'start_command'."""
    test_config = config_fixture.shell_command_before.config_unexpanded
    test_config = config.expand(test_config)

    assert test_config == config_fixture.shell_command_before.config_expanded()

    test_config = config.trickle(test_config)
    assert test_config == config_fixture.shell_command_before.config_after()


def test_in_session_scope(config_fixture: "ConfigTestData"):
    sconfig = ConfigReader._load(
        format="yaml", content=config_fixture.shell_command_before_session.before
    )

    config.validate_schema(sconfig)

    assert config.expand(sconfig) == sconfig
    assert config.expand(config.trickle(sconfig)) == ConfigReader._load(
        format="yaml", content=config_fixture.shell_command_before_session.expected
    )


def test_trickle_relative_start_directory(config_fixture: "ConfigTestData"):
    test_config = config.trickle(config_fixture.trickle.before)
    assert test_config == config_fixture.trickle.expected


def test_trickle_window_with_no_pane_config():
    test_yaml = """
    session_name: test_session
    windows:
    - window_name: test_1
      panes:
      - shell_command:
        - ls -l
    - window_name: test_no_panes
    """
    sconfig = ConfigReader._load(format="yaml", content=test_yaml)
    config.validate_schema(sconfig)

    assert config.expand(config.trickle(sconfig))["windows"][1]["panes"][0] == {
        "shell_command": []
    }


def test_expands_blank_panes(config_fixture: "ConfigTestData"):
    """Expand blank config into full form.

    Handle ``NoneType`` and 'blank'::

    # nothing, None, 'blank'
    'panes': [
        None,
        'blank'
    ]

    # should be blank
    'panes': [
        'shell_command': []
    ]

    Blank strings::

        panes: [
            ''
        ]

        # should output to:
        panes:
            'shell_command': ['']

    """
    yaml_config_file = EXAMPLE_PATH / "blank-panes.yaml"
    test_config = load_config(yaml_config_file)
    assert config.expand(test_config) == config_fixture.expand_blank.expected


def test_no_session_name():
    yaml_config = """
    - window_name: editor
      panes:
      shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    - window_name: logging
      automatic-rename: true
      panes:
      - shell_command:
      - htop
    """

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.matches(r'requires "session_name"')


def test_no_windows():
    yaml_config = """
    session_name: test session
    """

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.match(r'list of "windows"')


def test_no_window_name():
    yaml_config = """
    session_name: test session
    windows:
    - window_name: editor
      panes:
      shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    - automatic-rename: true
      panes:
      - shell_command:
      - htop
    """

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.matches('missing "window_name"')


def test_replaces_env_variables(monkeypatch):
    env_key = "TESTHEY92"
    env_val = "HEYO1"
    yaml_config = """
    start_directory: {TEST_VAR}/test
    shell_command_before: {TEST_VAR}/test2
    before_script: {TEST_VAR}/test3
    session_name: hi - {TEST_VAR}
    options:
        default-command: {TEST_VAR}/lol
    global_options:
        default-shell: {TEST_VAR}/moo
    windows:
    - window_name: editor
      panes:
      - shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    - window_name: logging @ {TEST_VAR}
      automatic-rename: true
      panes:
      - shell_command:
      - htop
    """.format(
        TEST_VAR="${%s}" % env_key
    )

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    monkeypatch.setenv(str(env_key), str(env_val))
    sconfig = config.expand(sconfig)
    assert "%s/test" % env_val == sconfig["start_directory"]
    assert (
        "%s/test2" % env_val
        in sconfig["shell_command_before"]["shell_command"][0]["cmd"]
    )
    assert "%s/test3" % env_val == sconfig["before_script"]
    assert "hi - %s" % env_val == sconfig["session_name"]
    assert "%s/moo" % env_val == sconfig["global_options"]["default-shell"]
    assert "%s/lol" % env_val == sconfig["options"]["default-command"]
    assert "logging @ %s" % env_val == sconfig["windows"][1]["window_name"]


def test_plugins():
    yaml_config = """
    session_name: test session
    plugins: tmuxp-plugin-one.plugin.TestPluginOne
    windows:
    - window_name: editor
      panes:
      shell_command:
      - tail -F /var/log/syslog
      start_directory: /var/log
    """

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    with pytest.raises(exc.ConfigError) as excinfo:
        config.validate_schema(sconfig)
        assert excinfo.matches("only supports list type")



from tmuxp.plugin import TmuxpPlugin


class MyTestTmuxpPlugin(TmuxpPlugin):
    def __init__(self, config):
        tmux_version = config.pop("tmux_version", None)
        libtmux_version = config.pop("libtmux_version", None)
        tmuxp_version = config.pop("tmuxp_version", None)

        TmuxpPlugin.__init__(self, **config)

        # WARNING! This should not be done in anything but a test
        if tmux_version:
            self.version_constraints["tmux"]["version"] = tmux_version
        if libtmux_version:
            self.version_constraints["libtmux"]["version"] = libtmux_version
        if tmuxp_version:
            self.version_constraints["tmuxp"]["version"] = tmuxp_version

        self._version_check()



from .. import utils as test_utils

teamocil_yaml_file = test_utils.get_config_file("config_teamocil/layouts.yaml")
teamocil_yaml = test_utils.read_config_file("config_teamocil/layouts.yaml")

teamocil_dict = {
    "two-windows": {
        "windows": [
            {
                "name": "foo",
                "clear": True,
                "root": "/foo",
                "layout": "tiled",
                "panes": [{"cmd": "echo 'foo'"}, {"cmd": "echo 'foo again'"}],
            },
            {
                "name": "bar",
                "root": "/bar",
                "splits": [
                    {
                        "cmd": ["echo 'bar'", "echo 'bar in an array'"],
                        "target": "bottom-right",
                    },
                    {"cmd": "echo 'bar again'", "focus": True, "width": 50},
                ],
            },
        ]
    },
    "two-windows-with-filters": {
        "windows": [
            {
                "name": "foo",
                "root": "/foo",
                "filters": {
                    "before": ["echo first before filter", "echo second before filter"],
                    "after": ["echo first after filter", "echo second after filter"],
                },
                "panes": [
                    {"cmd": "echo 'foo'"},
                    {"cmd": "echo 'foo again'", "width": 50},
                ],
            }
        ]
    },
    "two-windows-with-custom-command-options": {
        "windows": [
            {
                "name": "foo",
                "cmd_separator": "\n",
                "with_env_var": False,
                "clear": True,
                "root": "/foo",
                "layout": "tiled",
                "panes": [{"cmd": "echo 'foo'"}, {"cmd": "echo 'foo again'"}],
            },
            {
                "name": "bar",
                "cmd_separator": " && ",
                "with_env_var": True,
                "root": "/bar",
                "splits": [
                    {
                        "cmd": ["echo 'bar'", "echo 'bar in an array'"],
                        "target": "bottom-right",
                    },
                    {"cmd": "echo 'bar again'", "focus": True, "width": 50},
                ],
            },
        ]
    },
    "three-windows-within-a-session": {
        "session": {
            "name": "my awesome session",
            "windows": [
                {"name": "first window", "panes": [{"cmd": "echo 'foo'"}]},
                {"name": "second window", "panes": [{"cmd": "echo 'foo'"}]},
                {"name": "third window", "panes": [{"cmd": "echo 'foo'"}]},
            ],
        }
    },
}


two_windows = {
    "session_name": None,
    "windows": [
        {
            "window_name": "foo",
            "start_directory": "/foo",
            "clear": True,
            "layout": "tiled",
            "panes": [
                {"shell_command": "echo 'foo'"},
                {"shell_command": "echo 'foo again'"},
            ],
        },
        {
            "window_name": "bar",
            "start_directory": "/bar",
            "panes": [
                {
                    "shell_command": ["echo 'bar'", "echo 'bar in an array'"],
                    "target": "bottom-right",
                },
                {"shell_command": "echo 'bar again'", "focus": True},
            ],
        },
    ],
}

two_windows_with_filters = {
    "session_name": None,
    "windows": [
        {
            "window_name": "foo",
            "start_directory": "/foo",
            "shell_command_before": [
                "echo first before filter",
                "echo second before filter",
            ],
            "shell_command_after": [
                "echo first after filter",
                "echo second after filter",
            ],
            "panes": [
                {"shell_command": "echo 'foo'"},
                {"shell_command": "echo 'foo again'"},
            ],
        }
    ],
}

two_windows_with_custom_command_options = {
    "session_name": None,
    "windows": [
        {
            "window_name": "foo",
            "start_directory": "/foo",
            "clear": True,
            "layout": "tiled",
            "panes": [
                {"shell_command": "echo 'foo'"},
                {"shell_command": "echo 'foo again'"},
            ],
        },
        {
            "window_name": "bar",
            "start_directory": "/bar",
            "panes": [
                {
                    "shell_command": ["echo 'bar'", "echo 'bar in an array'"],
                    "target": "bottom-right",
                },
                {"shell_command": "echo 'bar again'", "focus": True},
            ],
        },
    ],
}

three_windows_within_a_session = {
    "session_name": "my awesome session",
    "windows": [
        {"window_name": "first window", "panes": [{"shell_command": "echo 'foo'"}]},
        {"window_name": "second window", "panes": [{"shell_command": "echo 'foo'"}]},
        {"window_name": "third window", "panes": [{"shell_command": "echo 'foo'"}]},
    ],
}



import os
import pathlib

import click

from tmuxp.config_reader import ConfigReader

from .utils import ConfigPath


@click.command(name="convert")
@click.option(
    "--yes", "-y", "confirmed", help='Auto confirms with "yes".', is_flag=True
)
@click.argument("config", type=ConfigPath(exists=True), nargs=1)
def command_convert(confirmed, config):
    """Convert a tmuxp config between JSON and YAML."""

    _, ext = os.path.splitext(config)
    ext = ext.lower()
    if ext == ".json":
        to_filetype = "yaml"
    elif ext in [".yaml", ".yml"]:
        to_filetype = "json"
    else:
        raise click.BadParameter(
            f"Unknown filetype: {ext} (valid: [.json, .yaml, .yml])"
        )

    configparser = ConfigReader.from_file(pathlib.Path(config))
    newfile = config.replace(ext, f".{to_filetype}")

    new_config = configparser.dump(format=to_filetype)

    if not confirmed:
        if click.confirm(f"convert to <{config}> to {to_filetype}?"):
            if click.confirm(f"Save config to {newfile}?"):
                confirmed = True

    if confirmed:
        buf = open(newfile, "w")
        buf.write(new_config)
        buf.close()
        print(f"New config saved to <{newfile}>.")



from .test_plugin_helpers import MyTestTmuxpPlugin


class TmuxVersionFailMinPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            "plugin_name": "tmux-min-version-fail",
            "tmux_min_version": "1.8",
            "tmux_version": "1.7",
        }
        MyTestTmuxpPlugin.__init__(self, config)


class TmuxVersionFailMaxPlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            "plugin_name": "tmux-max-version-fail",
            "tmux_max_version": "3.0",
            "tmux_version": "3.5",
        }
        MyTestTmuxpPlugin.__init__(self, config)


class TmuxVersionFailIncompatiblePlugin(MyTestTmuxpPlugin):
    def __init__(self):
        config = {
            "plugin_name": "tmux-incompatible-version-fail",
            "tmux_version_incompatible": ["2.3"],
            "tmux_version": "2.3",
        }

        MyTestTmuxpPlugin.__init__(self, config)



"""Conftest.py (root-level)

We keep this in root pytest fixtures in pytest's doctest plugin to be available, as well
as avoiding conftest.py from being included in the wheel.
"""
import logging
import os
import pathlib
import shutil
import typing as t

import pytest

from _pytest.doctest import DoctestItem

from libtmux.test import namer
from tests.fixtures import utils as test_utils

if t.TYPE_CHECKING:
    from libtmux.session import Session

logger = logging.getLogger(__name__)
USING_ZSH = "zsh" in os.getenv("SHELL", "")


@pytest.mark.skipif(USING_ZSH, reason="Using ZSH")
@pytest.fixture(autouse=USING_ZSH, scope="session")
def zshrc(user_path: pathlib.Path):
    """This quiets ZSH default message.

    Needs a startup file .zshenv, .zprofile, .zshrc, .zlogin.
    """
    p = user_path / ".zshrc"
    p.touch()
    return p


@pytest.fixture(autouse=True)
def home_path_default(monkeypatch: pytest.MonkeyPatch, user_path: pathlib.Path) -> None:
    monkeypatch.setenv("HOME", str(user_path))


@pytest.fixture(scope="function")
def monkeypatch_plugin_test_packages(monkeypatch):
    paths = [
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_bwb/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_bs/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_r/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_owc/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_awf/",
        "tests/fixtures/pluginsystem/plugins/tmuxp_test_plugin_fail/",
    ]
    for path in paths:
        monkeypatch.syspath_prepend(os.path.abspath(os.path.relpath(path)))


@pytest.fixture(scope="function")
def socket_name(request) -> str:
    return "tmuxp_test%s" % next(namer)


@pytest.fixture(autouse=True)
def add_doctest_fixtures(
    request: pytest.FixtureRequest,
    doctest_namespace: t.Dict[str, t.Any],
    tmp_path: pathlib.Path,
) -> None:
    if isinstance(request._pyfuncitem, DoctestItem) and shutil.which("tmux"):
        doctest_namespace["server"] = request.getfixturevalue("server")
        session: "Session" = request.getfixturevalue("session")
        doctest_namespace["session"] = session
        doctest_namespace["window"] = session.attached_window
        doctest_namespace["pane"] = session.attached_pane
        doctest_namespace["test_utils"] = test_utils
        doctest_namespace["tmp_path"] = tmp_path



"""Test for tmuxp tmuxinator configuration."""
import pytest

from tmuxp import config
from tmuxp.config_reader import ConfigReader

from .fixtures import config_tmuxinator as fixtures


@pytest.mark.parametrize(
    "tmuxinator_yaml,tmuxinator_dict,tmuxp_dict",
    [
        (
            fixtures.test1.tmuxinator_yaml,
            fixtures.test1.tmuxinator_dict,
            fixtures.test1.expected,
        ),
        (
            fixtures.test2.tmuxinator_yaml,
            fixtures.test2.tmuxinator_dict,
            fixtures.test2.expected,
        ),  # older vers use `tabs` instead of `windows`
        (
            fixtures.test3.tmuxinator_yaml,
            fixtures.test3.tmuxinator_dict,
            fixtures.test3.expected,
        ),  # Test importing <spec/fixtures/sample.yml>
    ],
)
def test_config_to_dict(tmuxinator_yaml, tmuxinator_dict, tmuxp_dict):
    yaml_to_dict = ConfigReader._load(format="yaml", content=tmuxinator_yaml)
    assert yaml_to_dict == tmuxinator_dict

    assert config.import_tmuxinator(tmuxinator_dict) == tmuxp_dict

    config.validate_schema(config.import_tmuxinator(tmuxinator_dict))



"""Tests for freezing tmux sessions with tmuxp."""
import time

from tmuxp import config
from tmuxp.config_reader import ConfigReader
from tmuxp.workspacebuilder import WorkspaceBuilder, freeze

from .fixtures import utils as test_utils


def test_freeze_config(session):
    session_config = ConfigReader._from_file(
        test_utils.get_config_file("workspacefreezer/sampleconfig.yaml")
    )

    builder = WorkspaceBuilder(sconf=session_config)
    builder.build(session=session)
    assert session == builder.session

    time.sleep(0.50)

    session = session
    new_config = freeze(session)

    config.validate_schema(new_config)

    # These should dump without an error
    ConfigReader._dump(format="json", content=new_config)
    ConfigReader._dump(format="yaml", content=new_config)

    # Inline configs should also dump without an error
    compact_config = config.inline(new_config)

    ConfigReader._dump(format="json", content=compact_config)
    ConfigReader._dump(format="yaml", content=compact_config)



"""Test for tmuxp command line interface."""
import json
import os
import pathlib
import typing as t

import pytest

import click
from click.testing import CliRunner
from pytest_mock import MockerFixture

import libtmux
from libtmux.common import has_lt_version
from libtmux.exc import LibTmuxException
from tmuxp import cli, config, exc
from tmuxp.cli.debug_info import command_debug_info
from tmuxp.cli.import_config import get_teamocil_dir, get_tmuxinator_dir
from tmuxp.cli.load import (
    _load_append_windows_to_current_session,
    _load_attached,
    _reattach,
    load_plugins,
    load_workspace,
)
from tmuxp.cli.ls import command_ls
from tmuxp.cli.utils import (
    ConfigPath,
    _validate_choices,
    get_abs_path,
    get_config_dir,
    is_pure_name,
    scan_config,
)
from tmuxp.config_reader import ConfigReader
from tmuxp.workspacebuilder import WorkspaceBuilder

from .constants import FIXTURE_PATH
from .fixtures import utils as test_utils

if t.TYPE_CHECKING:
    from libtmux.server import Server


def test_creates_config_dir_not_exists(tmp_path: pathlib.Path):
    """cli.startup() creates config dir if not exists."""

    cli.startup(tmp_path)
    assert os.path.exists(tmp_path)


def test_in_dir_from_config_dir(tmp_path: pathlib.Path):
    """config.in_dir() finds configs config dir."""

    cli.startup(tmp_path)
    yaml_config = tmp_path / "myconfig.yaml"
    yaml_config.touch()
    json_config = tmp_path / "myconfig.json"
    json_config.touch()
    configs_found = config.in_dir(tmp_path)

    assert len(configs_found) == 2


def test_ignore_non_configs_from_current_dir(tmp_path: pathlib.Path):
    """cli.in_dir() ignore non-config from config dir."""

    cli.startup(tmp_path)

    junk_config = tmp_path / "myconfig.psd"
    junk_config.touch()
    conf = tmp_path / "watmyconfig.json"
    conf.touch()
    configs_found = config.in_dir(tmp_path)
    assert len(configs_found) == 1


def test_get_configs_cwd(tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    """config.in_cwd() find config in shell current working directory."""

    confdir = tmp_path / "tmuxpconf2"
    confdir.mkdir()

    monkeypatch.chdir(confdir)
    config1 = open(".tmuxp.json", "w+b")
    config1.close()

    configs_found = config.in_cwd()
    assert len(configs_found) == 1
    assert ".tmuxp.json" in configs_found


@pytest.mark.parametrize(
    "path,expect",
    [
        (".", False),
        ("./", False),
        ("", False),
        (".tmuxp.yaml", False),
        ("../.tmuxp.yaml", False),
        ("../", False),
        ("/hello/world", False),
        ("~/.tmuxp/hey", False),
        ("~/work/c/tmux/", False),
        ("~/work/c/tmux/.tmuxp.yaml", False),
        ("myproject", True),
    ],
)
def test_is_pure_name(path, expect):
    assert is_pure_name(path) == expect


"""
    scans for .tmuxp.{yaml,yml,json} in directory, returns first result
    log warning if multiple found:

    - current directory: ., ./, noarg
    - relative to cwd directory: ../, ./hello/, hello/, ./hello/
    - absolute directory: /path/to/dir, /path/to/dir/, ~/
    - no path, no ext, config_dir: projectname, tmuxp

    load file directly -

    - no directory (cwd): .tmuxp.yaml
    - relative to cwd: ../.tmuxp.yaml, ./hello/.tmuxp.yaml
    - absolute path: /path/to/file.yaml, ~/path/to/file/.tmuxp.yaml

    Any case where file is not found should return error.
"""


@pytest.fixture
def homedir(tmp_path: pathlib.Path):
    home = tmp_path / "home"
    home.mkdir()
    return home


@pytest.fixture
def configdir(homedir):
    conf = homedir / ".tmuxp"
    conf.mkdir()
    return conf


@pytest.fixture
def projectdir(homedir):
    proj = homedir / "work" / "project"
    proj.mkdir(parents=True)
    return proj


def test_tmuxp_configdir_env_var(tmp_path: pathlib.Path, monkeypatch):
    monkeypatch.setenv("TMUXP_CONFIGDIR", str(tmp_path))

    assert get_config_dir() == str(tmp_path)


def test_tmuxp_configdir_xdg_config_dir(tmp_path: pathlib.Path, monkeypatch):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    tmux_dir = tmp_path / "tmuxp"
    tmux_dir.mkdir()

    assert get_config_dir() == str(tmux_dir)


def test_resolve_dot(
    tmp_path: pathlib.Path,
    homedir: pathlib.Path,
    configdir: pathlib.Path,
    projectdir: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setenv("HOME", str(homedir))
    monkeypatch.setenv("XDG_CONFIG_HOME", str(homedir / ".config"))

    tmuxp_conf_path = projectdir / ".tmuxp.yaml"
    tmuxp_conf_path.touch()
    user_config_name = "myconfig"
    user_config = configdir / f"{user_config_name}.yaml"
    user_config.touch()

    project_config = tmuxp_conf_path

    monkeypatch.chdir(projectdir)

    expect = str(project_config)
    assert scan_config(".") == expect
    assert scan_config("./") == expect
    assert scan_config("") == expect
    assert scan_config("../project") == expect
    assert scan_config("../project/") == expect
    assert scan_config(".tmuxp.yaml") == expect
    assert scan_config("../../.tmuxp/%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("myconfig") == str(user_config)
    assert scan_config("~/.tmuxp/myconfig.yaml") == str(user_config)

    with pytest.raises(Exception):
        scan_config(".tmuxp.json")
    with pytest.raises(Exception):
        scan_config(".tmuxp.ini")
    with pytest.raises(Exception):
        scan_config("../")
    with pytest.raises(Exception):
        scan_config("mooooooo")

    monkeypatch.chdir(homedir)

    expect = str(project_config)
    assert scan_config("work/project") == expect
    assert scan_config("work/project/") == expect
    assert scan_config("./work/project") == expect
    assert scan_config("./work/project/") == expect
    assert scan_config(".tmuxp/%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("./.tmuxp/%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("myconfig") == str(user_config)
    assert scan_config("~/.tmuxp/myconfig.yaml") == str(user_config)

    with pytest.raises(Exception):
        scan_config("")
    with pytest.raises(Exception):
        scan_config(".")
    with pytest.raises(Exception):
        scan_config(".tmuxp.yaml")
    with pytest.raises(Exception):
        scan_config("../")
    with pytest.raises(Exception):
        scan_config("mooooooo")

    monkeypatch.chdir(configdir)

    expect = str(project_config)
    assert scan_config("../work/project") == expect
    assert scan_config("../../home/work/project") == expect
    assert scan_config("../work/project/") == expect
    assert scan_config("%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("./%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("myconfig") == str(user_config)
    assert scan_config("~/.tmuxp/myconfig.yaml") == str(user_config)

    with pytest.raises(Exception):
        scan_config("")
    with pytest.raises(Exception):
        scan_config(".")
    with pytest.raises(Exception):
        scan_config(".tmuxp.yaml")
    with pytest.raises(Exception):
        scan_config("../")
    with pytest.raises(Exception):
        scan_config("mooooooo")

    monkeypatch.chdir(tmp_path)

    expect = str(project_config)
    assert scan_config("home/work/project") == expect
    assert scan_config("./home/work/project/") == expect
    assert scan_config("home/.tmuxp/%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("./home/.tmuxp/%s.yaml" % user_config_name) == str(user_config)
    assert scan_config("myconfig") == str(user_config)
    assert scan_config("~/.tmuxp/myconfig.yaml") == str(user_config)

    with pytest.raises(Exception):
        scan_config("")
    with pytest.raises(Exception):
        scan_config(".")
    with pytest.raises(Exception):
        scan_config(".tmuxp.yaml")
    with pytest.raises(Exception):
        scan_config("../")
    with pytest.raises(Exception):
        scan_config("mooooooo")


def test_scan_config_arg(
    homedir, configdir, projectdir, monkeypatch: pytest.MonkeyPatch
):
    runner = CliRunner()

    @click.command()
    @click.argument("config", type=ConfigPath(exists=True), nargs=-1)
    def config_cmd(config):
        click.echo(config)

    monkeypatch.setenv("HOME", str(homedir))
    tmuxp_config_path = projectdir / ".tmuxp.yaml"
    tmuxp_config_path.touch()
    user_config_name = "myconfig"
    user_config = configdir / f"{user_config_name}.yaml"
    user_config.touch()

    project_config = projectdir / ".tmuxp.yaml"

    def check_cmd(config_arg):
        return runner.invoke(config_cmd, [config_arg]).output

    monkeypatch.chdir(projectdir)
    expect = str(project_config)
    assert expect in check_cmd(".")
    assert expect in check_cmd("./")
    assert expect in check_cmd("")
    assert expect in check_cmd("../project")
    assert expect in check_cmd("../project/")
    assert expect in check_cmd(".tmuxp.yaml")
    assert str(user_config) in check_cmd("../../.tmuxp/%s.yaml" % user_config_name)
    assert user_config.stem in check_cmd("myconfig")
    assert str(user_config) in check_cmd("~/.tmuxp/myconfig.yaml")

    assert "file not found" in check_cmd(".tmuxp.json")
    assert "file not found" in check_cmd(".tmuxp.ini")
    assert "No tmuxp files found" in check_cmd("../")
    assert "config not found in config dir" in check_cmd("moo")


def test_load_workspace(server, monkeypatch):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv("TMUX", raising=False)
    session_file = FIXTURE_PATH / "workspacebuilder" / "two_pane.yaml"

    # open it detached
    session = load_workspace(
        session_file, socket_name=server.socket_name, detached=True
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == "sampleconfig"


def test_load_workspace_named_session(server, monkeypatch):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv("TMUX", raising=False)
    session_file = FIXTURE_PATH / "workspacebuilder" / "two_pane.yaml"

    # open it detached
    session = load_workspace(
        session_file,
        socket_name=server.socket_name,
        new_session_name="tmuxp-new",
        detached=True,
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == "tmuxp-new"


@pytest.mark.skipif(
    has_lt_version("2.1"), reason="exact session name matches only tmux >= 2.1"
)
def test_load_workspace_name_match_regression_252(
    tmp_path: pathlib.Path, server, monkeypatch
):
    monkeypatch.delenv("TMUX", raising=False)
    session_file = FIXTURE_PATH / "workspacebuilder" / "two_pane.yaml"

    # open it detached
    session = load_workspace(
        session_file, socket_name=server.socket_name, detached=True
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == "sampleconfig"

    projfile = tmp_path / "simple.yaml"

    projfile.write_text(
        """
session_name: sampleconfi
start_directory: './'
windows:
- panes:
    - echo 'hey'""",
        encoding="utf-8",
    )

    # open it detached
    session = load_workspace(
        str(projfile), socket_name=server.socket_name, detached=True
    )
    assert session.name == "sampleconfi"


def test_load_symlinked_workspace(server, tmp_path, monkeypatch):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv("TMUX", raising=False)

    realtemp = tmp_path / "myrealtemp"
    realtemp.mkdir()
    linktemp = tmp_path / "symlinktemp"
    linktemp.symlink_to(realtemp)
    projfile = linktemp / "simple.yaml"

    projfile.write_text(
        """
session_name: samplesimple
start_directory: './'
windows:
- panes:
    - echo 'hey'""",
        encoding="utf-8",
    )

    # open it detached
    session = load_workspace(
        str(projfile), socket_name=server.socket_name, detached=True
    )
    pane = session.attached_window.attached_pane

    assert isinstance(session, libtmux.Session)
    assert session.name == "samplesimple"
    assert pane.current_path == str(realtemp)


def test_regression_00132_session_name_with_dots(
    tmp_path: pathlib.Path, server, session
):
    yaml_config = FIXTURE_PATH / "workspacebuilder" / "regression_00132_dots.yaml"
    cli_args = [str(yaml_config)]
    inputs: t.List[str] = []
    runner = CliRunner()
    result = runner.invoke(
        cli.command_load, cli_args, input="".join(inputs), standalone_mode=False
    )
    assert result.exception
    assert isinstance(result.exception, libtmux.exc.BadSessionName)


@pytest.mark.parametrize("cli_args", [(["load", "."]), (["load", ".tmuxp.yaml"])])
def test_load_zsh_autotitle_warning(cli_args, tmp_path, monkeypatch):
    # create dummy tmuxp yaml so we don't get yelled at
    yaml_config = tmp_path / ".tmuxp.yaml"
    yaml_config.touch()
    oh_my_zsh_path = tmp_path / ".oh-my-zsh"
    oh_my_zsh_path.mkdir()
    monkeypatch.setenv("HOME", str(tmp_path))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    monkeypatch.delenv("DISABLE_AUTO_TITLE", raising=False)
    monkeypatch.setenv("SHELL", "zsh")
    result = runner.invoke(cli.cli, cli_args)
    assert "Please set" in result.output

    monkeypatch.setenv("DISABLE_AUTO_TITLE", "false")
    result = runner.invoke(cli.cli, cli_args)
    assert "Please set" in result.output

    monkeypatch.setenv("DISABLE_AUTO_TITLE", "true")
    result = runner.invoke(cli.cli, cli_args)
    assert "Please set" not in result.output

    monkeypatch.delenv("DISABLE_AUTO_TITLE", raising=False)
    monkeypatch.setenv("SHELL", "sh")
    result = runner.invoke(cli.cli, cli_args)
    assert "Please set" not in result.output


@pytest.mark.parametrize(
    "cli_args",
    [
        (["load", ".", "--log-file", "log.txt"]),
    ],
)
def test_load_log_file(cli_args, tmp_path, monkeypatch):
    # create dummy tmuxp yaml that breaks to prevent actually loading tmux
    tmuxp_config_path = tmp_path / ".tmuxp.yaml"
    tmuxp_config_path.write_text(
        """
session_name: hello
        """,
        encoding="utf-8",
    )
    oh_my_zsh_path = tmp_path / ".oh-my-zsh"
    oh_my_zsh_path.mkdir()
    monkeypatch.setenv("HOME", str(tmp_path))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    # If autoconfirm (-y) no need to prompt y
    input_args = "y\ny\n" if "-y" not in cli_args else ""

    result = runner.invoke(cli.cli, cli_args, input=input_args)
    log_file_path = tmp_path / "log.txt"
    assert "Loading" in log_file_path.open().read()
    assert result is not None


@pytest.mark.parametrize("cli_cmd", ["shell", ("shell", "--pdb")])
@pytest.mark.parametrize(
    "cli_args,inputs,env,expected_output",
    [
        (
            ["-L{SOCKET_NAME}", "-c", "print(str(server.socket_name))"],
            [],
            {},
            "{SERVER_SOCKET_NAME}",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "{SESSION_NAME}",
                "-c",
                "print(session.name)",
            ],
            [],
            {},
            "{SESSION_NAME}",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "{SESSION_NAME}",
                "{WINDOW_NAME}",
                "-c",
                "print(server.has_session(session.name))",
            ],
            [],
            {},
            "True",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "{SESSION_NAME}",
                "{WINDOW_NAME}",
                "-c",
                "print(window.name)",
            ],
            [],
            {},
            "{WINDOW_NAME}",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "{SESSION_NAME}",
                "{WINDOW_NAME}",
                "-c",
                "print(pane.id)",
            ],
            [],
            {},
            "{PANE_ID}",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "-c",
                "print(pane.id)",
            ],
            [],
            {"TMUX_PANE": "{PANE_ID}"},
            "{PANE_ID}",
        ),
    ],
)
def test_shell(
    cli_cmd,
    cli_args,
    inputs,
    expected_output,
    env,
    tmp_path,
    monkeypatch,
    server,
    session,
):
    monkeypatch.setenv("HOME", str(tmp_path))
    window_name = "my_window"
    window = session.new_window(window_name=window_name)
    window.split_window()

    template_ctx = dict(
        SOCKET_NAME=server.socket_name,
        SOCKET_PATH=server.socket_path,
        SESSION_NAME=session.name,
        WINDOW_NAME=window_name,
        PANE_ID=window.attached_pane.id,
        SERVER_SOCKET_NAME=server.socket_name,
    )

    cli_cmd = list(cli_cmd) if isinstance(cli_cmd, (list, tuple)) else [cli_cmd]
    cli_args = cli_cmd + [cli_arg.format(**template_ctx) for cli_arg in cli_args]

    for k, v in env.items():
        monkeypatch.setenv(k, v.format(**template_ctx))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    result = runner.invoke(
        cli.cli, cli_args, input="".join(inputs), catch_exceptions=False
    )
    assert expected_output.format(**template_ctx) in result.output


@pytest.mark.parametrize(
    "cli_cmd",
    [
        "shell",
        ("shell", "--pdb"),
    ],
)
@pytest.mark.parametrize(
    "cli_args,inputs,env,template_ctx,exception,message",
    [
        (
            ["-LDoesNotExist", "-c", "print(str(server.socket_name))"],
            [],
            {},
            {},
            LibTmuxException,
            r".*DoesNotExist.*",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "nonexistant_session",
                "-c",
                "print(str(server.socket_name))",
            ],
            [],
            {},
            {"session_name": "nonexistant_session"},
            exc.TmuxpException,
            "Session not found: nonexistant_session",
        ),
        (
            [
                "-L{SOCKET_NAME}",
                "{SESSION_NAME}",
                "nonexistant_window",
                "-c",
                "print(str(server.socket_name))",
            ],
            [],
            {},
            {"window_name": "nonexistant_window"},
            exc.TmuxpException,
            "Window not found: {WINDOW_NAME}",
        ),
    ],
)
def test_shell_target_missing(
    cli_cmd,
    cli_args,
    inputs,
    env,
    template_ctx,
    exception,
    message,
    tmp_path,
    monkeypatch,
    socket_name,
    server,
    session,
):
    monkeypatch.setenv("HOME", str(tmp_path))
    window_name = "my_window"
    window = session.new_window(window_name=window_name)
    window.split_window()

    template_ctx = dict(
        SOCKET_NAME=server.socket_name,
        SOCKET_PATH=server.socket_path,
        SESSION_NAME=session.name,
        WINDOW_NAME=template_ctx.get("window_name", window_name),
        PANE_ID=template_ctx.get("pane_id"),
        SERVER_SOCKET_NAME=server.socket_name,
    )
    cli_cmd = list(cli_cmd) if isinstance(cli_cmd, (list, tuple)) else [cli_cmd]
    cli_args = cli_cmd + [cli_arg.format(**template_ctx) for cli_arg in cli_args]

    for k, v in env.items():
        monkeypatch.setenv(k, v.format(**template_ctx))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    if exception is not None:
        with pytest.raises(exception, match=message.format(**template_ctx)):
            result = runner.invoke(
                cli.cli, cli_args, input="".join(inputs), catch_exceptions=False
            )
    else:
        result = runner.invoke(
            cli.cli, cli_args, input="".join(inputs), catch_exceptions=False
        )
        assert message.format(**template_ctx) in result.output


@pytest.mark.parametrize(
    "cli_cmd",
    [
        # 'shell',
        # ('shell', '--pdb'),
        ("shell", "--code"),
        # ('shell', '--bpython'),
        # ('shell', '--ptipython'),
        # ('shell', '--ptpython'),
        # ('shell', '--ipython'),
    ],
)
@pytest.mark.parametrize(
    "cli_args,inputs,env,message",
    [
        (
            [
                "-L{SOCKET_NAME}",
            ],
            [],
            {},
            "(InteractiveConsole)",
        ),
        (
            [
                "-L{SOCKET_NAME}",
            ],
            [],
            {"PANE_ID": "{PANE_ID}"},
            "(InteractiveConsole)",
        ),
    ],
)
def test_shell_plus(
    cli_cmd,
    cli_args,
    inputs,
    env,
    message,
    tmp_path,
    monkeypatch,
    server,
    session,
):
    monkeypatch.setenv("HOME", str(tmp_path))
    window_name = "my_window"
    window = session.new_window(window_name=window_name)
    window.split_window()

    template_ctx = dict(
        SOCKET_NAME=server.socket_name,
        SOCKET_PATH=server.socket_path,
        SESSION_NAME=session.name,
        WINDOW_NAME=window_name,
        PANE_ID=window.attached_pane.id,
        SERVER_SOCKET_NAME=server.socket_name,
    )

    cli_cmd = list(cli_cmd) if isinstance(cli_cmd, (list, tuple)) else [cli_cmd]
    cli_args = cli_cmd + [cli_arg.format(**template_ctx) for cli_arg in cli_args]

    for k, v in env.items():
        monkeypatch.setenv(k, v.format(**template_ctx))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    result = runner.invoke(
        cli.cli, cli_args, input="".join(inputs), catch_exceptions=True
    )
    assert message.format(**template_ctx) in result.output


@pytest.mark.parametrize(
    "cli_args",
    [
        (["convert", "."]),
        (["convert", ".tmuxp.yaml"]),
        (["convert", ".tmuxp.yaml", "-y"]),
        (["convert", ".tmuxp.yml"]),
        (["convert", ".tmuxp.yml", "-y"]),
    ],
)
def test_convert(cli_args, tmp_path, monkeypatch):
    # create dummy tmuxp yaml so we don't get yelled at
    filename = cli_args[1]
    if filename == ".":
        filename = ".tmuxp.yaml"
    file_ext = filename.rsplit(".", 1)[-1]
    assert file_ext in ["yaml", "yml"], file_ext
    config_file_path = tmp_path / filename
    config_file_path.write_text("\nsession_name: hello\n", encoding="utf-8")
    oh_my_zsh_path = tmp_path / ".oh-my-zsh"
    oh_my_zsh_path.mkdir()
    monkeypatch.setenv("HOME", str(tmp_path))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    # If autoconfirm (-y) no need to prompt y
    input_args = "y\ny\n" if "-y" not in cli_args else ""

    runner.invoke(cli.cli, cli_args, input=input_args)
    tmuxp_json = tmp_path / ".tmuxp.json"
    assert tmuxp_json.exists()
    assert tmuxp_json.open().read() == json.dumps({"session_name": "hello"}, indent=2)


@pytest.mark.parametrize(
    "cli_args",
    [
        (["convert", "."]),
        (["convert", ".tmuxp.json"]),
        (["convert", ".tmuxp.json", "-y"]),
    ],
)
def test_convert_json(cli_args, tmp_path, monkeypatch):
    # create dummy tmuxp yaml so we don't get yelled at
    json_config = tmp_path / ".tmuxp.json"
    json_config.write_text('{"session_name": "hello"}', encoding="utf-8")
    oh_my_zsh_path = tmp_path / ".oh-my-zsh"
    oh_my_zsh_path.mkdir()
    monkeypatch.setenv("HOME", str(tmp_path))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()

    # If autoconfirm (-y) no need to prompt y
    input_args = "y\ny\n" if "-y" not in cli_args else ""

    runner.invoke(cli.cli, cli_args, input=input_args)
    tmuxp_yaml = tmp_path / ".tmuxp.yaml"
    assert tmuxp_yaml.exists()
    assert tmuxp_yaml.open().read() == "session_name: hello\n"


@pytest.mark.parametrize("cli_args", [(["import"])])
def test_import(cli_args, monkeypatch):
    runner = CliRunner()

    result = runner.invoke(cli.cli, cli_args)
    assert "tmuxinator" in result.output
    assert "teamocil" in result.output


@pytest.mark.parametrize(
    "cli_args",
    [
        (["--help"]),
        (["-h"]),
    ],
)
def test_help(cli_args, monkeypatch):
    runner = CliRunner()

    result = runner.invoke(cli.cli, cli_args)
    assert "Usage: cli [OPTIONS] COMMAND [ARGS]..." in result.output


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ["import", "teamocil", "./.teamocil/config.yaml"],
            ["\n", "y\n", "./la.yaml\n", "y\n"],
        ),
        (
            ["import", "teamocil", "./.teamocil/config.yaml"],
            ["\n", "y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
        ),
        (
            ["import", "teamocil", "config"],
            ["\n", "y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
        ),
    ],
)
def test_import_teamocil(cli_args, inputs, tmp_path, monkeypatch):
    teamocil_config = test_utils.read_config_file("config_teamocil/test4.yaml")

    teamocil_path = tmp_path / ".teamocil"
    teamocil_path.mkdir()

    teamocil_config_path = teamocil_path / "config.yaml"
    teamocil_config_path.write_text(teamocil_config, encoding="utf-8")

    exists_yaml = tmp_path / "exists.yaml"
    exists_yaml.touch()

    monkeypatch.setenv("HOME", str(tmp_path))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()
    runner.invoke(cli.cli, cli_args, input="".join(inputs))

    new_config_yaml = tmp_path / "la.yaml"
    assert new_config_yaml.exists()


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ["import", "tmuxinator", "./.tmuxinator/config.yaml"],
            ["\n", "y\n", "./la.yaml\n", "y\n"],
        ),
        (
            ["import", "tmuxinator", "./.tmuxinator/config.yaml"],
            ["\n", "y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
        ),
        (
            ["import", "tmuxinator", "config"],
            ["\n", "y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
        ),
    ],
)
def test_import_tmuxinator(cli_args, inputs, tmp_path, monkeypatch):
    tmuxinator_config = test_utils.read_config_file("config_tmuxinator/test3.yaml")

    tmuxinator_path = tmp_path / ".tmuxinator"
    tmuxinator_path.mkdir()

    tmuxinator_config_path = tmuxinator_path / "config.yaml"
    tmuxinator_config_path.write_text(tmuxinator_config, encoding="utf-8")

    exists_yaml = tmp_path / "exists.yaml"
    exists_yaml.touch()

    monkeypatch.setenv("HOME", str(tmp_path))

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()
    out = runner.invoke(cli.cli, cli_args, input="".join(inputs))
    print(out.output)
    new_config_yaml = tmp_path / "la.yaml"
    assert new_config_yaml.exists()


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (["freeze", "myfrozensession"], ["y\n", "./la.yaml\n", "y\n"]),
        (  # Exists
            ["freeze", "myfrozensession"],
            ["y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"],
        ),
        (  # Imply current session if not entered
            ["freeze"],
            ["y\n", "./la.yaml\n", "y\n"],
        ),
        (["freeze"], ["y\n", "./exists.yaml\n", "./la.yaml\n", "y\n"]),  # Exists
    ],
)
def test_freeze(server, cli_args, inputs, tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))
    exists_yaml = tmp_path / "exists.yaml"
    exists_yaml.touch()

    server.new_session(session_name="myfirstsession")
    server.new_session(session_name="myfrozensession")

    # Assign an active pane to the session
    second_session = server.list_sessions()[1]
    first_pane_on_second_session_id = second_session.list_windows()[0].list_panes()[0][
        "pane_id"
    ]
    monkeypatch.setenv("TMUX_PANE", first_pane_on_second_session_id)

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()
    # Use tmux server (socket name) used in the test
    cli_args = cli_args + ["-L", server.socket_name]
    out = runner.invoke(cli.cli, cli_args, input="".join(inputs))
    print(out.output)

    yaml_config_path = tmp_path / "la.yaml"
    assert yaml_config_path.exists()

    yaml_config = yaml_config_path.open().read()
    frozen_config = ConfigReader._load(format="yaml", content=yaml_config)

    assert frozen_config["session_name"] == "myfrozensession"


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (  # Overwrite
            ["freeze", "mysession", "--force"],
            ["\n", "y\n", "./exists.yaml\n", "y\n"],
        ),
        (  # Imply current session if not entered
            ["freeze", "--force"],
            ["\n", "y\n", "./exists.yaml\n", "y\n"],
        ),
    ],
)
def test_freeze_overwrite(server, cli_args, inputs, tmp_path, monkeypatch):
    monkeypatch.setenv("HOME", str(tmp_path))
    exists_yaml = tmp_path / "exists.yaml"
    exists_yaml.touch()

    server.new_session(session_name="mysession")

    monkeypatch.chdir(tmp_path)
    runner = CliRunner()
    # Use tmux server (socket name) used in the test
    cli_args = cli_args + ["-L", server.socket_name]
    out = runner.invoke(cli.cli, cli_args, input="".join(inputs))
    print(out.output)

    yaml_config_path = tmp_path / "exists.yaml"
    assert yaml_config_path.exists()


def test_get_abs_path(tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    expect = str(tmp_path)
    monkeypatch.chdir(tmp_path)
    get_abs_path("../") == os.path.dirname(expect)
    get_abs_path(".") == expect
    get_abs_path("./") == expect
    get_abs_path(expect) == expect


def test_get_tmuxinator_dir(monkeypatch):
    assert get_tmuxinator_dir() == os.path.expanduser("~/.tmuxinator/")

    monkeypatch.setenv("HOME", "/moo")
    assert get_tmuxinator_dir() == "/moo/.tmuxinator/"
    assert get_tmuxinator_dir() == os.path.expanduser("~/.tmuxinator/")


def test_get_teamocil_dir(monkeypatch: pytest.MonkeyPatch):
    assert get_teamocil_dir() == os.path.expanduser("~/.teamocil/")

    monkeypatch.setenv("HOME", "/moo")
    assert get_teamocil_dir() == "/moo/.teamocil/"
    assert get_teamocil_dir() == os.path.expanduser("~/.teamocil/")


def test_validate_choices():
    validate = _validate_choices(["choice1", "choice2"])

    assert validate("choice1")
    assert validate("choice2")

    with pytest.raises(click.BadParameter):
        assert validate("choice3")


def test_pass_config_dir_ClickPath(
    tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch
):

    configdir = tmp_path / "myconfigdir"
    configdir.mkdir()
    user_config_name = "myconfig"
    user_config = configdir / f"{user_config_name}.yaml"
    user_config.touch()

    expect = str(user_config)

    runner = CliRunner()

    @click.command()
    @click.argument(
        "config",
        type=ConfigPath(exists=True, config_dir=(str(configdir))),
        nargs=-1,
    )
    def config_cmd(config):
        click.echo(config)

    def check_cmd(config_arg):
        return runner.invoke(config_cmd, [config_arg]).output

    monkeypatch.chdir(configdir)

    assert expect in check_cmd("myconfig")
    assert expect in check_cmd("myconfig.yaml")
    assert expect in check_cmd("./myconfig.yaml")
    assert str(user_config) in check_cmd(str(configdir / "myconfig.yaml"))

    assert "file not found" in check_cmd(".tmuxp.json")


def test_ls_cli(monkeypatch: pytest.MonkeyPatch, tmp_path: pathlib.Path):
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path / ".config"))

    filenames = [
        ".git/",
        ".gitignore/",
        "session_1.yaml",
        "session_2.yaml",
        "session_3.json",
        "session_4.txt",
    ]

    # should ignore:
    # - directories should be ignored
    # - extensions not covered in VALID_CONFIG_DIR_FILE_EXTENSIONS
    ignored_filenames = [".git/", ".gitignore/", "session_4.txt"]
    stems = [os.path.splitext(f)[0] for f in filenames if f not in ignored_filenames]

    for filename in filenames:
        location = tmp_path / f".tmuxp/{filename}"
        if filename.endswith("/"):
            location.mkdir(parents=True)
        else:
            location.touch()

    runner = CliRunner()
    cli_output = runner.invoke(command_ls).output
    assert cli_output == "\n".join(stems) + "\n"


def test_load_plugins(monkeypatch_plugin_test_packages):
    from tmuxp_test_plugin_bwb.plugin import PluginBeforeWorkspaceBuilder

    plugins_config = test_utils.read_config_file("workspacebuilder/plugin_bwb.yaml")

    sconfig = ConfigReader._load(format="yaml", content=plugins_config)
    sconfig = config.expand(sconfig)

    plugins = load_plugins(sconfig)

    assert len(plugins) == 1

    test_plugin_class_types = [
        PluginBeforeWorkspaceBuilder().__class__,
    ]
    for plugin in plugins:
        assert plugin.__class__ in test_plugin_class_types


@pytest.mark.skip("Not sure how to clean up the tmux session this makes")
@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ["load", "tests/fixtures/workspacebuilder/plugin_versions_fail.yaml"],
            ["y\n"],
        )
    ],
)
def test_load_plugins_version_fail_skip(
    monkeypatch_plugin_test_packages, cli_args, inputs
):
    runner = CliRunner()

    results = runner.invoke(cli.cli, cli_args, input="".join(inputs))
    assert "[Loading]" in results.output


@pytest.mark.parametrize(
    "cli_args,inputs",
    [
        (
            ["load", "tests/fixtures/workspacebuilder/plugin_versions_fail.yaml"],
            ["n\n"],
        )
    ],
)
def test_load_plugins_version_fail_no_skip(
    monkeypatch_plugin_test_packages, cli_args, inputs
):
    runner = CliRunner()

    results = runner.invoke(cli.cli, cli_args, input="".join(inputs))
    assert "[Not Skipping]" in results.output


@pytest.mark.parametrize(
    "cli_args", [(["load", "tests/fixtures/workspacebuilder/plugin_missing_fail.yaml"])]
)
def test_load_plugins_plugin_missing(monkeypatch_plugin_test_packages, cli_args):
    runner = CliRunner()

    results = runner.invoke(cli.cli, cli_args)
    assert "[Plugin Error]" in results.output


def test_plugin_system_before_script(
    monkeypatch_plugin_test_packages, server, monkeypatch
):
    # this is an implementation test. Since this testsuite may be ran within
    # a tmux session by the developer himself, delete the TMUX variable
    # temporarily.
    monkeypatch.delenv("TMUX", raising=False)
    session_file = FIXTURE_PATH / "workspacebuilder" / "plugin_bs.yaml"

    # open it detached
    session = load_workspace(
        session_file, socket_name=server.socket_name, detached=True
    )

    assert isinstance(session, libtmux.Session)
    assert session.name == "plugin_test_bs"


def test_reattach_plugins(monkeypatch_plugin_test_packages, server):
    config_plugins = test_utils.read_config_file("workspacebuilder/plugin_r.yaml")

    sconfig = ConfigReader._load(format="yaml", content=config_plugins)
    sconfig = config.expand(sconfig)

    # open it detached
    builder = WorkspaceBuilder(
        sconf=sconfig, plugins=load_plugins(sconfig), server=server
    )
    builder.build()

    try:
        _reattach(builder)
    except libtmux.exc.LibTmuxException:
        pass

    proc = builder.session.cmd("display-message", "-p", "'#S'")

    assert proc.stdout[0] == "'plugin_test_r'"


def test_load_attached(
    server: "Server", monkeypatch: pytest.MonkeyPatch, mocker: MockerFixture
) -> None:
    # Load a session and attach from outside tmux
    monkeypatch.delenv("TMUX", raising=False)

    attach_session_mock = mocker.patch("libtmux.session.Session.attach_session")
    attach_session_mock.return_value.stderr = None

    yaml_config = test_utils.read_config_file("workspacebuilder/two_pane.yaml")
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, False)

    assert attach_session_mock.call_count == 1


def test_load_attached_detached(
    server: "Server", monkeypatch: pytest.MonkeyPatch, mocker: MockerFixture
) -> None:
    # Load a session but don't attach
    monkeypatch.delenv("TMUX", raising=False)

    attach_session_mock = mocker.patch("libtmux.session.Session.attach_session")
    attach_session_mock.return_value.stderr = None

    yaml_config = test_utils.read_config_file("workspacebuilder/two_pane.yaml")
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, True)

    assert attach_session_mock.call_count == 0


def test_load_attached_within_tmux(
    server: "Server", monkeypatch: pytest.MonkeyPatch, mocker: MockerFixture
) -> None:
    # Load a session and attach from within tmux
    monkeypatch.setenv("TMUX", "/tmp/tmux-1234/default,123,0")

    switch_client_mock = mocker.patch("libtmux.session.Session.switch_client")
    switch_client_mock.return_value.stderr = None

    yaml_config = test_utils.read_config_file("workspacebuilder/two_pane.yaml")
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, False)

    assert switch_client_mock.call_count == 1


def test_load_attached_within_tmux_detached(
    server: "Server", monkeypatch: pytest.MonkeyPatch, mocker: MockerFixture
) -> None:
    # Load a session and attach from within tmux
    monkeypatch.setenv("TMUX", "/tmp/tmux-1234/default,123,0")

    switch_client_mock = mocker.patch("libtmux.session.Session.switch_client")
    switch_client_mock.return_value.stderr = None

    yaml_config = test_utils.read_config_file("workspacebuilder/two_pane.yaml")
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    _load_attached(builder, True)

    assert switch_client_mock.call_count == 1


def test_load_append_windows_to_current_session(server, monkeypatch):
    yaml_config = test_utils.read_config_file("workspacebuilder/two_pane.yaml")
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.list_sessions()) == 1
    assert len(server._list_windows()) == 3

    # Assign an active pane to the session
    monkeypatch.setenv("TMUX_PANE", server._list_panes()[0]["pane_id"])

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    _load_append_windows_to_current_session(builder)

    assert len(server.list_sessions()) == 1
    assert len(server._list_windows()) == 6


def test_debug_info_cli(monkeypatch, tmp_path: pathlib.Path):
    monkeypatch.setenv("SHELL", "/bin/bash")

    runner = CliRunner()
    cli_output = runner.invoke(command_debug_info).output
    assert "environment" in cli_output
    assert "python version" in cli_output
    assert "system PATH" in cli_output
    assert "tmux version" in cli_output
    assert "libtmux version" in cli_output
    assert "tmuxp version" in cli_output
    assert "tmux path" in cli_output
    assert "tmuxp path" in cli_output
    assert "shell" in cli_output
    assert "tmux session" in cli_output
    assert "tmux windows" in cli_output
    assert "tmux panes" in cli_output
    assert "tmux global options" in cli_output
    assert "tmux window options" in cli_output



"""Test for tmuxp workspacebuilder."""
import os
import pathlib
import textwrap
import time
import typing as t

import pytest

import libtmux
from libtmux.common import has_gte_version, has_lt_version
from libtmux.test import retry_until, temp_session
from libtmux.window import Window
from tmuxp import config, exc
from tmuxp.cli.load import load_plugins
from tmuxp.config_reader import ConfigReader
from tmuxp.workspacebuilder import WorkspaceBuilder

from .constants import EXAMPLE_PATH, FIXTURE_PATH
from .fixtures import utils as test_utils

if t.TYPE_CHECKING:
    from libtmux.server import Server


def test_split_windows(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/two_pane.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(session._windows) == window_count
    for w, wconf in builder.iter_create_windows(session):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout("tiled")  # fix glitch with pane size
            p = p
            assert len(session._windows) == window_count
        assert isinstance(w, Window)

        assert len(session._windows) == window_count
        window_count += 1


def test_split_windows_three_pane(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/three_pane.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(session._windows) == window_count
    for w, wconf in builder.iter_create_windows(session):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout("tiled")  # fix glitch with pane size
            p = p
            assert len(session._windows) == window_count
        assert isinstance(w, Window)

        assert len(session._windows) == window_count
        window_count += 1
        w.set_window_option("main-pane-height", 50)
        w.select_layout(wconf["layout"])


def test_focus_pane_index(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/focus_and_pane.yaml")
    )
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    builder.build(session=session)

    assert session.attached_window.name == "focused window"

    pane_base_index = int(
        session.attached_window.show_window_option("pane-base-index", g=True)
    )

    if not pane_base_index:
        pane_base_index = 0
    else:
        pane_base_index = int(pane_base_index)

    # get the pane index for each pane
    pane_base_indexes = []
    for pane in session.attached_window.panes:
        pane_base_indexes.append(int(pane.index))

    pane_indexes_should_be = [pane_base_index + x for x in range(0, 3)]
    assert pane_indexes_should_be == pane_base_indexes

    w = session.attached_window

    assert w.name != "man"

    pane_path = "/usr"
    p = None

    def f():
        nonlocal p
        p = w.attached_pane
        p.server._update_panes()
        return p.current_path == pane_path

    assert retry_until(f)

    assert p.current_path == pane_path

    proc = session.cmd("show-option", "-gv", "base-index")
    base_index = int(proc.stdout[0])
    session.server._update_windows()

    window3 = session.find_where({"window_index": str(base_index + 2)})
    assert isinstance(window3, Window)

    p = None
    pane_path = "/"

    def f():
        nonlocal p
        p = window3.attached_pane
        p.server._update_panes()
        return p.current_path == pane_path

    assert retry_until(f)

    assert p.current_path == pane_path


@pytest.mark.skip(
    reason="""
Test needs to be rewritten, assertion not reliable across platforms
and CI. See https://github.com/tmux-python/tmuxp/issues/310.
    """.strip()
)
def test_suppress_history(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/suppress_history.yaml")
    )
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    inHistoryWindow = session.find_where({"window_name": "inHistory"})
    isMissingWindow = session.find_where({"window_name": "isMissing"})

    def assertHistory(cmd, hist):
        return "inHistory" in cmd and cmd.endswith(hist)

    def assertIsMissing(cmd, hist):
        return "isMissing" in cmd and not cmd.endswith(hist)

    for w, window_name, assertCase in [
        (inHistoryWindow, "inHistory", assertHistory),
        (isMissingWindow, "isMissing", assertIsMissing),
    ]:
        assert w.name == window_name
        w.select_window()
        p = w.attached_pane
        p.select_pane()

        # Print the last-in-history command in the pane
        p.cmd("send-keys", " fc -ln -1")
        p.cmd("send-keys", "Enter")

        buffer_name = "test"
        sent_cmd = None

        def f():
            # from v0.7.4 libtmux session.cmd adds target -t self.id by default
            # show-buffer doesn't accept -t, use global cmd.

            # Get the contents of the pane
            p.cmd("capture-pane", "-b", buffer_name)

            captured_pane = session.server.cmd("show-buffer", "-b", buffer_name)
            session.server.cmd("delete-buffer", "-b", buffer_name)

            # Parse the sent and last-in-history commands
            sent_cmd = captured_pane.stdout[0].strip()
            history_cmd = captured_pane.stdout[-2].strip()

            return assertCase(sent_cmd, history_cmd)

        assert retry_until(f), f"Unknown sent command: [{sent_cmd}] in {assertCase}"


def test_session_options(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/session_options.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert "/bin/sh" in session.show_option("default-shell")
    assert "/bin/sh" in session.show_option("default-command")


def test_global_options(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/global_options.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert "top" in session.show_option("status-position", _global=True)
    assert 493 == session.show_option("repeat-time", _global=True)


def test_global_session_env_options(session, monkeypatch):
    visual_silence = "on"
    monkeypatch.setenv("VISUAL_SILENCE", str(visual_silence))
    repeat_time = 738
    monkeypatch.setenv("REPEAT_TIME", str(repeat_time))
    main_pane_height = 8
    monkeypatch.setenv("MAIN_PANE_HEIGHT", str(main_pane_height))

    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/env_var_options.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert visual_silence in session.show_option("visual-silence", _global=True)
    assert repeat_time == session.show_option("repeat-time")
    assert main_pane_height == session.attached_window.show_window_option(
        "main-pane-height"
    )


def test_window_options(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/window_options.yaml")
    )
    sconfig = config.expand(sconfig)

    if has_gte_version("2.3"):
        sconfig["windows"][0]["options"]["pane-border-format"] = " #P "

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(session._windows) == window_count
    for w, wconf in builder.iter_create_windows(session):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout("tiled")  # fix glitch with pane size
            p = p
            assert len(session._windows) == window_count
        assert isinstance(w, Window)
        assert w.show_window_option("main-pane-height") == 5
        if has_gte_version("2.3"):
            assert w.show_window_option("pane-border-format") == " #P "

        assert len(session._windows) == window_count
        window_count += 1
        w.select_layout(wconf["layout"])


@pytest.mark.flaky(reruns=5)
def test_window_options_after(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/window_options_after.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    def assert_last_line(p, s):
        def f():
            pane_out = p.cmd("capture-pane", "-p", "-J").stdout
            while not pane_out[-1].strip():  # delete trailing lines tmux 1.8
                pane_out.pop()
            return len(pane_out) > 1 and pane_out[-2].strip() == s

        # Print output for easier debugging if assertion fails
        return retry_until(f, raises=False)

    for i, pane in enumerate(session.attached_window.panes):
        assert assert_last_line(
            pane, str(i)
        ), "Initial command did not execute properly/" + str(i)
        pane.cmd("send-keys", "Up")  # Will repeat echo
        pane.enter()  # in each iteration
        assert assert_last_line(
            pane, str(i)
        ), "Repeated command did not execute properly/" + str(i)

    session.cmd("send-keys", " echo moo")
    session.cmd("send-keys", "Enter")

    for pane in session.attached_window.panes:
        assert assert_last_line(
            pane, "moo"
        ), "Synchronized command did not execute properly"


def test_window_shell(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/window_shell.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    for w, wconf in builder.iter_create_windows(session):
        if "window_shell" in wconf:
            assert wconf["window_shell"] == "top"

        def f():
            session.server._update_windows()
            return w["window_name"] != "top"

        retry_until(f)

        assert w.name != "top"


def test_environment_variables(session):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/environment_vars.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session)

    assert session.getenv("FOO") == "BAR"
    assert session.getenv("PATH") == "/tmp"


def test_automatic_rename_option(session):
    """With option automatic-rename: on."""
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/window_automatic_rename.yaml")
    )

    # This should be a command guaranteed to be terminal name across systems
    portable_command = sconfig["windows"][0]["panes"][0]["shell_command"][0]["cmd"]
    # If a command is like "man ls", get the command base name, "ls"
    if " " in portable_command:
        portable_command = portable_command.split(" ")[0]

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(session._windows) == window_count
    for w, wconf in builder.iter_create_windows(session):

        for p in builder.iter_create_panes(w, wconf):
            w.select_layout("tiled")  # fix glitch with pane size
            p = p
            assert len(session._windows), window_count
        assert isinstance(w, Window)
        assert w.show_window_option("automatic-rename") == "on"

        assert len(session._windows) == window_count

        window_count += 1
        w.select_layout(wconf["layout"])

    assert session.name != "tmuxp"
    w = session.windows[0]

    def check_window_name_mismatch() -> bool:
        session.server._update_windows()
        return w.name != portable_command

    assert retry_until(check_window_name_mismatch, 2, interval=0.25)

    pane_base_index = w.show_window_option("pane-base-index", g=True)
    w.select_pane(pane_base_index)

    def check_window_name_match() -> bool:
        session.server._update_windows()
        return w.name == portable_command

    assert retry_until(
        check_window_name_match, 2, interval=0.25
    ), f"Window name {w.name} should be {portable_command}"

    w.select_pane("-D")

    assert retry_until(check_window_name_mismatch, 2, interval=0.25)


def test_blank_pane_count(session):
    """:todo: Verify blank panes of various types build into workspaces."""
    yaml_config_file = EXAMPLE_PATH / "blank-panes.yaml"
    test_config = ConfigReader._from_file(yaml_config_file)

    test_config = config.expand(test_config)
    builder = WorkspaceBuilder(sconf=test_config)
    builder.build(session=session)

    assert session == builder.session

    window1 = session.find_where({"window_name": "Blank pane test"})
    assert len(window1._panes) == 3

    window2 = session.find_where({"window_name": "More blank panes"})
    assert len(window2._panes) == 3

    window3 = session.find_where({"window_name": "Empty string (return)"})
    assert len(window3._panes) == 3

    window4 = session.find_where({"window_name": "Blank with options"})
    assert len(window4._panes) == 2


def test_start_directory(session, tmp_path: pathlib.Path):
    test_dir = tmp_path / "foo bar"
    test_dir.mkdir()

    yaml_config = test_utils.read_config_file("workspacebuilder/start_directory.yaml")
    test_config = yaml_config.format(TEST_DIR=test_dir)

    sconfig = ConfigReader._load(format="yaml", content=test_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert session == builder.session
    dirs = ["/usr/bin", "/dev", str(test_dir), "/usr", "/usr"]

    for path, window in zip(dirs, session.windows):
        for p in window.panes:

            def f():
                p.server._update_panes()
                pane_path = p.current_path
                return path in pane_path or pane_path == path

            # handle case with OS X adding /private/ to /tmp/ paths
            assert retry_until(f)


def test_start_directory_relative(session, tmp_path: pathlib.Path):
    """Same as above test, but with relative start directory, mimicking
    loading it from a location of project file. Like::

    $ tmuxp load ~/workspace/myproject/.tmuxp.yaml

    instead of::

    $ cd ~/workspace/myproject/.tmuxp.yaml
    $ tmuxp load .

    """
    yaml_config = test_utils.read_config_file(
        "workspacebuilder/start_directory_relative.yaml"
    )

    test_dir = tmp_path / "foo bar"
    test_dir.mkdir()
    config_dir = tmp_path / "testRelConfigDir"
    config_dir.mkdir()

    test_config = yaml_config.format(TEST_DIR=test_dir)
    sconfig = ConfigReader._load(format="yaml", content=test_config)
    # the second argument of os.getcwd() mimics the behavior
    # the CLI loader will do, but it passes in the config file's location.
    sconfig = config.expand(sconfig, config_dir)

    sconfig = config.trickle(sconfig)

    assert os.path.exists(config_dir)
    assert os.path.exists(test_dir)
    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert session == builder.session

    dirs = ["/usr/bin", "/dev", str(test_dir), str(config_dir), str(config_dir)]

    for path, window in zip(dirs, session.windows):
        for p in window.panes:

            def f():
                p.server._update_panes()
                # Handle case where directories resolve to /private/ in OSX
                pane_path = p.current_path
                return path in pane_path or pane_path == path

            assert retry_until(f)


@pytest.mark.skipif(
    has_lt_version("3.2a"), reason="needs format introduced in tmux >= 3.2a"
)
def test_start_directory_sets_session_path(server):
    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/start_directory_session_path.yaml")
    )
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    session = builder.session
    expected = "{0}|/usr".format(session.id)

    cmd = server.cmd("list-sessions", "-F", "#{session_id}|#{session_path}")
    assert expected in cmd.stdout


def test_pane_order(session):
    """Pane ordering based on position in config and ``pane_index``.

    Regression test for https://github.com/tmux-python/tmuxp/issues/15.
    """
    yaml_config = test_utils.read_config_file(
        "workspacebuilder/pane_ordering.yaml"
    ).format(HOME=os.path.realpath(os.path.expanduser("~")))

    # test order of `panes` (and pane_index) above against pane_dirs
    pane_paths = [
        "/usr/bin",
        "/usr",
        "/etc",
        os.path.realpath(os.path.expanduser("~")),
    ]

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    window_count = len(session._windows)  # current window count
    assert len(session._windows) == window_count

    for w, wconf in builder.iter_create_windows(session):
        for p in builder.iter_create_panes(w, wconf):
            w.select_layout("tiled")  # fix glitch with pane size
            assert len(session._windows) == window_count

        assert isinstance(w, Window)

        assert len(session._windows) == window_count
        window_count += 1

    for w in session.windows:
        pane_base_index = w.show_window_option("pane-base-index", g=True)
        for p_index, p in enumerate(w.list_panes(), start=pane_base_index):
            assert int(p_index) == int(p.index)

            # pane-base-index start at base-index, pane_paths always start
            # at 0 since python list.
            pane_path = pane_paths[p_index - pane_base_index]

            def f():
                p.server._update_panes()
                return p.current_path == pane_path

            assert retry_until(f)


def test_window_index(session):
    proc = session.cmd("show-option", "-gv", "base-index")
    base_index = int(proc.stdout[0])
    name_index_map = {"zero": 0 + base_index, "one": 1 + base_index, "five": 5}

    sconfig = ConfigReader._from_file(
        test_utils.get_config_file("workspacebuilder/window_index.yaml")
    )
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    for window, _ in builder.iter_create_windows(session):
        expected_index = name_index_map[window["window_name"]]
        assert int(window["window_index"]) == expected_index


def test_before_load_throw_error_if_retcode_error(server):
    config_script_fails = test_utils.read_config_file(
        "workspacebuilder/config_script_fails.yaml"
    )
    yaml_config = config_script_fails.format(
        script_failed=FIXTURE_PATH / "script_failed.sh",
    )

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as sess:
        session_name = sess.name

        with pytest.raises(exc.BeforeLoadScriptError):
            builder.build(session=sess)

        result = server.has_session(session_name)
        assert not result, "Kills session if before_script exits with errcode"


def test_before_load_throw_error_if_file_not_exists(server):
    config_script_not_exists = test_utils.read_config_file(
        "workspacebuilder/config_script_not_exists.yaml"
    )
    yaml_config = config_script_not_exists.format(
        script_not_exists=FIXTURE_PATH / "script_not_exists.sh",
    )
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as session:
        session_name = session.name
        temp_session_exists = server.has_session(session.name)
        assert temp_session_exists
        with pytest.raises((exc.BeforeLoadScriptNotExists, OSError)) as excinfo:
            builder.build(session=session)
            excinfo.match(r"No such file or directory")
        result = server.has_session(session_name)
        assert not result, "Kills session if before_script doesn't exist"


def test_before_load_true_if_test_passes(server):
    config_script_completes = test_utils.read_config_file(
        "workspacebuilder/config_script_completes.yaml"
    )
    script_complete_sh = FIXTURE_PATH / "script_complete.sh"
    assert script_complete_sh.exists()

    yaml_config = config_script_completes.format(script_complete=script_complete_sh)
    sconfig = ConfigReader._load(format="yaml", content=yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as session:
        builder.build(session=session)


def test_before_load_true_if_test_passes_with_args(server):
    config_script_completes = test_utils.read_config_file(
        "workspacebuilder/config_script_completes.yaml"
    )
    script_complete_sh = FIXTURE_PATH / "script_complete.sh"
    assert script_complete_sh.exists()

    yaml_config = config_script_completes.format(script_complete=script_complete_sh)

    sconfig = ConfigReader._load(format="yaml", content=yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)

    with temp_session(server) as session:
        builder.build(session=session)


def test_plugin_system_before_workspace_builder(
    monkeypatch_plugin_test_packages, session
):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/plugin_bwb.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd("display-message", "-p", "'#S'")
    assert proc.stdout[0] == "'plugin_test_bwb'"


def test_plugin_system_on_window_create(monkeypatch_plugin_test_packages, session):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/plugin_owc.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd("display-message", "-p", "'#W'")
    assert proc.stdout[0] == "'plugin_test_owc'"


def test_plugin_system_after_window_finished(monkeypatch_plugin_test_packages, session):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/plugin_awf.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd("display-message", "-p", "'#W'")
    assert proc.stdout[0] == "'plugin_test_awf'"


def test_plugin_system_on_window_create_multiple_windows(session):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file(
            "workspacebuilder/plugin_owc_multiple_windows.yaml"
        )
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd("list-windows", "-F", "'#W'")
    assert "'plugin_test_owc_mw'" in proc.stdout
    assert "'plugin_test_owc_mw_2'" in proc.stdout


def test_plugin_system_after_window_finished_multiple_windows(
    monkeypatch_plugin_test_packages, session
):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file(
            "workspacebuilder/plugin_awf_multiple_windows.yaml"
        )
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    proc = session.cmd("list-windows", "-F", "'#W'")
    assert "'plugin_test_awf_mw'" in proc.stdout
    assert "'plugin_test_awf_mw_2'" in proc.stdout


def test_plugin_system_multiple_plugins(monkeypatch_plugin_test_packages, session):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/plugin_multiple_plugins.yaml")
    )
    sconfig = config.expand(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig, plugins=load_plugins(sconfig))
    assert len(builder.plugins) > 0

    builder.build(session=session)

    # Drop through to the before_script plugin hook
    proc = session.cmd("display-message", "-p", "'#S'")
    assert proc.stdout[0] == "'plugin_test_bwb'"

    # Drop through to the after_window_finished. This won't succeed
    # unless on_window_create succeeds because of how the test plugin
    # override methods are currently written
    proc = session.cmd("display-message", "-p", "'#W'")
    assert proc.stdout[0] == "'mp_test_awf'"


def test_load_configs_same_session(server):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/three_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.sessions) == 1
    assert len(server.sessions[0]._windows) == 3

    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/two_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()
    assert len(server.sessions) == 2
    assert len(server.sessions[1]._windows) == 2

    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/two_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build(server.sessions[1], True)

    assert len(server.sessions) == 2
    assert len(server.sessions[1]._windows) == 4


def test_load_configs_separate_sessions(server):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/three_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.sessions) == 1
    assert len(server.sessions[0]._windows) == 3

    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/two_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.sessions) == 2
    assert len(server.sessions[0]._windows) == 3
    assert len(server.sessions[1]._windows) == 2


def test_find_current_active_pane(server, monkeypatch):
    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/three_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    sconfig = ConfigReader._from_file(
        path=test_utils.get_config_file("workspacebuilder/two_windows.yaml")
    )

    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    assert len(server.list_sessions()) == 2

    # Assign an active pane to the session
    second_session = server.list_sessions()[1]
    first_pane_on_second_session_id = second_session.list_windows()[0].list_panes()[0][
        "pane_id"
    ]
    monkeypatch.setenv("TMUX_PANE", first_pane_on_second_session_id)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    assert builder.find_current_attached_session() == second_session


@pytest.mark.parametrize(
    "yaml,output,should_see",
    [
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command: echo "___$((1 + 3))___"
    enter: false
    """
            ),
            "___4___",
            False,
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - echo "___$((1 + 3))___"
    enter: false
    """
            ),
            "___4___",
            False,
        ],
        [
            textwrap.dedent(
                """
session_name: Should execute
windows:
- panes:
  - shell_command: echo "___$((1 + 3))___"
  """
            ),
            "___4___",
            True,
        ],
        [
            textwrap.dedent(
                """
session_name: Should execute
windows:
- panes:
  - shell_command:
    - echo "___$((1 + 3))___"
  """
            ),
            "___4___",
            True,
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - cmd: echo "___$((1 + 3))___"
      enter: false
    """
            ),
            "___4___",
            False,
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - cmd: echo "___$((1 + 3))___"
      enter: false
    """
            ),
            "___4___",
            False,
        ],
        [
            textwrap.dedent(
                """
session_name: Should execute
windows:
- panes:
  - shell_command: echo "___$((1 + 3))___"
  """
            ),
            "___4___",
            True,
        ],
        [
            textwrap.dedent(
                """
session_name: Should execute
windows:
- panes:
  - shell_command:
    - cmd: echo "other command"
    - cmd: echo "___$((1 + 3))___"
  """
            ),
            "___4___",
            True,
        ],
    ],
    ids=[
        "pane_enter_false_shortform",
        "pane_enter_false_longform",
        "pane_enter_default_shortform",
        "pane_enter_default_longform",
        "pane_command_enter_false_shortform",
        "pane_command_enter_false_longform",
        "pane_command_enter_default_shortform",
        "pane_command_enter_default_longform",
    ],
)
def test_load_workspace_enter(
    tmp_path: pathlib.Path,
    server: libtmux.Server,
    monkeypatch: pytest.MonkeyPatch,
    yaml,
    output,
    should_see,
):
    yaml_config = tmp_path / "simple.yaml"
    yaml_config.write_text(yaml, encoding="utf-8")
    sconfig = ConfigReader._from_file(yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)
    builder = WorkspaceBuilder(sconf=sconfig, server=server)
    builder.build()

    session = builder.session
    pane = session.attached_pane

    def fn():
        captured_pane = "\n".join(pane.capture_pane())

        if should_see:
            return output in captured_pane
        else:
            return output not in captured_pane

    assert retry_until(
        fn, 1
    ), f'Should{" " if should_see else "not "} output in captured pane'


@pytest.mark.parametrize(
    "yaml,sleep,output",
    [
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - cmd: echo "___$((1 + 5))___"
      sleep_before: .15
    - cmd: echo "___$((1 + 3))___"
      sleep_before: .35
    """
            ),
            0.5,
            "___4___",
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - cmd: echo "___$((1 + 5))___"
      sleep_before: 1
    - cmd: echo "___$((1 + 3))___"
      sleep_before: .25
    """
            ),
            1.25,
            "___4___",
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - cmd: echo "___$((1 + 3))___"
    sleep_before: .5
    """
            ),
            0.5,
            "___4___",
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
windows:
- panes:
  - shell_command:
    - cmd: echo "___$((1 + 3))___"
    sleep_before: 1
    """
            ),
            1,
            "___4___",
        ],
        [
            textwrap.dedent(
                """
session_name: Should not execute
shell_command_before:
  - cmd: echo "sleeping before"
    sleep_before: .5
windows:
- panes:
  - echo "___$((1 + 3))___"
    """
            ),
            0.5,
            "___4___",
        ],
    ],
    ids=[
        "command_level_sleep_shortform",
        "command_level_pane_sleep_longform",
        "pane_sleep_shortform",
        "pane_sleep_longform",
        "shell_before_before_command_level",
    ],
)
@pytest.mark.flaky(reruns=3)
def test_load_workspace_sleep(
    tmp_path: pathlib.Path,
    server: libtmux.Server,
    monkeypatch: pytest.MonkeyPatch,
    yaml,
    sleep: int,
    output,
):
    yaml_config = tmp_path / "simple.yaml"
    yaml_config.write_text(yaml, encoding="utf-8")
    sconfig = ConfigReader._from_file(yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)
    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    t = time.process_time()

    builder.build()
    time.sleep(0.5)
    session = builder.session
    pane = session.attached_pane

    while (time.process_time() - t) * 1000 < sleep:
        captured_pane = "\n".join(pane.capture_pane())

        assert output not in captured_pane
        time.sleep(0.1)
    captured_pane = "\n".join(pane.capture_pane())
    assert output in captured_pane


def test_first_pane_start_directory(session, tmp_path: pathlib.Path):
    yaml_config = test_utils.get_config_file(
        "workspacebuilder/first_pane_start_directory.yaml"
    )

    sconfig = ConfigReader._from_file(yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert session == builder.session
    dirs = ["/usr", "/etc"]

    assert session.windows
    window = session.windows[0]
    for path, p in zip(dirs, window.panes):

        def f():
            p.server._update_panes()
            pane_path = p.current_path
            return path in pane_path or pane_path == path

        # handle case with OS X adding /private/ to /tmp/ paths
        assert retry_until(f)


@pytest.mark.skipif(
    has_lt_version("2.9"), reason="needs option introduced in tmux >= 2.9"
)
def test_layout_main_horizontal(session):
    yaml_config = test_utils.get_config_file("workspacebuilder/three_pane.yaml")
    sconfig = ConfigReader._from_file(path=yaml_config)

    builder = WorkspaceBuilder(sconf=sconfig)
    builder.build(session=session)

    assert session.windows
    window = session.windows[0]

    assert len(window.panes) == 3
    main_horizontal_pane, *panes = window.panes

    def height(p):
        return int(p._info["pane_height"])

    def width(p):
        return int(p._info["pane_width"])

    main_horizontal_pane_height = height(main_horizontal_pane)
    pane_heights = [height(pane) for pane in panes]
    # TODO: When libtmux has new pane formatters added, use that to detect top / bottom
    assert all(
        main_horizontal_pane_height != pane_height for pane_height in pane_heights
    ), "The top row should not be the same size as the bottom row (even though it can)"
    assert all(
        pane_heights[0] == pane_height for pane_height in pane_heights
    ), "The bottom row should be uniform height"
    assert width(main_horizontal_pane) > width(panes[0])

    def is_almost_equal(x, y):
        return abs(x - y) <= 1

    assert is_almost_equal(height(panes[0]), height(panes[1]))
    assert is_almost_equal(width(panes[0]), width(panes[1]))


class DefaultSizeNamespaceFixture(t.NamedTuple):
    test_id: str
    TMUXP_DEFAULT_SIZE: t.Optional[str]
    raises: bool
    confoverrides: t.Dict[str, t.Any]


DEFAULT_SIZE_FIXTURES = [
    DefaultSizeNamespaceFixture(
        test_id="default-behavior",
        TMUXP_DEFAULT_SIZE=None,
        raises=False,
        confoverrides={},
    ),
    DefaultSizeNamespaceFixture(
        test_id="v1.13.1 default-size-breaks",
        TMUXP_DEFAULT_SIZE=None,
        raises=True,
        confoverrides={"options": {"default-size": "80x24"}},
    ),
    DefaultSizeNamespaceFixture(
        test_id="v1.13.1-option-workaround",
        TMUXP_DEFAULT_SIZE=None,
        raises=False,
        confoverrides={"options": {"default-size": "800x600"}},
    ),
]


@pytest.mark.parametrize(
    DefaultSizeNamespaceFixture._fields,
    DEFAULT_SIZE_FIXTURES,
    ids=[f.test_id for f in DEFAULT_SIZE_FIXTURES],
)
@pytest.mark.skipif(has_lt_version("2.9"), reason="default-size only applies there")
def test_issue_800_default_size_many_windows(
    server: "Server",
    monkeypatch: pytest.MonkeyPatch,
    test_id: str,
    TMUXP_DEFAULT_SIZE: t.Optional[str],
    raises: bool,
    confoverrides: t.Dict[str, t.Any],
) -> None:
    """Recreate default-size issue.

    v1.13.1 added a default-size, but this can break building workspaces with
    a lot of panes.

    See also: https://github.com/tmux-python/tmuxp/issues/800
    """
    yaml_config = test_utils.get_config_file(
        "regressions/issue_800_default_size_many_windows.yaml"
    )

    sconfig = ConfigReader._from_file(yaml_config)
    sconfig = config.expand(sconfig)
    sconfig = config.trickle(sconfig)

    if isinstance(confoverrides, dict):
        for k, v in confoverrides.items():
            sconfig[k] = v

    if TMUXP_DEFAULT_SIZE is not None:
        monkeypatch.setenv("TMUXP_DEFAULT_SIZE", TMUXP_DEFAULT_SIZE)

    builder = WorkspaceBuilder(sconf=sconfig, server=server)

    if raises:
        with pytest.raises(Exception):
            builder.build()

        builder.session.kill_session()

        with pytest.raises(libtmux.exc.LibTmuxException, match="no space for new pane"):
            builder.build()
        return

    builder.build()
    assert len(server.list_sessions()) == 1



import os
import pathlib
import sys

import click

from tmuxp.config_reader import ConfigReader

from .. import config
from .utils import ConfigPath, _validate_choices, get_abs_path, tmuxp_echo


def get_tmuxinator_dir():
    """
    Return tmuxinator configuration directory.

    Checks for ``TMUXINATOR_CONFIG`` environmental variable.

    Returns
    -------
    str :
        absolute path to tmuxinator config directory

    See Also
    --------
    :meth:`tmuxp.config.import_tmuxinator`
    """
    if "TMUXINATOR_CONFIG" in os.environ:
        return os.path.expanduser(os.environ["TMUXINATOR_CONFIG"])

    return os.path.expanduser("~/.tmuxinator/")


def get_teamocil_dir():
    """
    Return teamocil configuration directory.

    Returns
    -------
    str :
        absolute path to teamocil config directory

    See Also
    --------
    :meth:`tmuxp.config.import_teamocil`
    """
    return os.path.expanduser("~/.teamocil/")


def _resolve_path_no_overwrite(config):
    path = get_abs_path(config)
    if os.path.exists(path):
        raise click.exceptions.UsageError("%s exists. Pick a new filename." % path)
    return path


@click.group(name="import")
def command_import():
    """Import a teamocil/tmuxinator config."""


def import_config(configfile, importfunc):
    existing_config = ConfigReader._from_file(pathlib.Path(configfile))
    new_config = ConfigReader(importfunc(existing_config))

    config_format = click.prompt(
        "Convert to", value_proc=_validate_choices(["yaml", "json"]), default="yaml"
    )

    if config_format == "yaml":
        new_config = new_config.dump("yaml", indent=2, default_flow_style=False)
    elif config_format == "json":
        new_config = new_config.dump("json", indent=2)
    else:
        sys.exit("Unknown config format.")

    tmuxp_echo(
        new_config + "---------------------------------------------------------------"
        "\n"
        "Configuration import does its best to convert files.\n"
    )
    if click.confirm(
        "The new config *WILL* require adjusting afterwards. Save config?"
    ):
        dest = None
        while not dest:
            dest_path = click.prompt(
                "Save to [%s]" % os.getcwd(), value_proc=_resolve_path_no_overwrite
            )

            # dest = dest_prompt
            if click.confirm("Save to %s?" % dest_path):
                dest = dest_path

        buf = open(dest, "w")
        buf.write(new_config)
        buf.close()

        tmuxp_echo("Saved to %s." % dest)
    else:
        tmuxp_echo(
            "tmuxp has examples in JSON and YAML format at "
            "<http://tmuxp.git-pull.com/examples.html>\n"
            "View tmuxp docs at <http://tmuxp.git-pull.com/>"
        )
        sys.exit()


@command_import.command(
    name="tmuxinator", short_help="Convert and import a tmuxinator config."
)
@click.argument(
    "configfile", type=ConfigPath(exists=True, config_dir=get_tmuxinator_dir), nargs=1
)
def command_import_tmuxinator(configfile):
    """Convert a tmuxinator config from CONFIGFILE to tmuxp format and import
    it into tmuxp."""
    import_config(configfile, config.import_tmuxinator)


@click.command(name="convert")
@click.option(
    "--yes", "-y", "confirmed", help='Auto confirms with "yes".', is_flag=True
)
@click.argument("config", type=ConfigPath(exists=True), nargs=1)
def command_convert(confirmed, config):
    """Convert a tmuxp config between JSON and YAML."""

    _, ext = os.path.splitext(config)
    ext = ext.lower()
    if ext == ".json":
        to_filetype = "yaml"
    elif ext in [".yaml", ".yml"]:
        to_filetype = "json"
    else:
        raise click.BadParameter(
            f"Unknown filetype: {ext} (valid: [.json, .yaml, .yml])"
        )

    configparser = ConfigReader.from_file(config)
    newfile = config.replace(ext, ".%s" % to_filetype)

    export_kwargs = {"default_flow_style": False} if to_filetype == "yaml" else {}
    new_config = configparser.dump(format=to_filetype, indent=2, **export_kwargs)

    if not confirmed:
        if click.confirm(f"convert to <{config}> to {to_filetype}?"):
            if click.confirm("Save config to %s?" % newfile):
                confirmed = True

    if confirmed:
        buf = open(newfile, "w")
        buf.write(new_config)
        buf.close()
        print("New config saved to <%s>." % newfile)


@command_import.command(
    name="teamocil", short_help="Convert and import a teamocil config."
)
@click.argument(
    "configfile", type=ConfigPath(exists=True, config_dir=get_teamocil_dir), nargs=1
)
def command_import_teamocil(configfile):
    """Convert a teamocil config from CONFIGFILE to tmuxp format and import
    it into tmuxp."""

    import_config(configfile, config.import_teamocil)



"""Command line tool for managing tmux workspaces and tmuxp configurations.

tmuxp.cli
~~~~~~~~~

"""
import importlib
import logging
import os
import pathlib
import shutil
import sys
from typing import List

import click

from libtmux.common import has_gte_version
from libtmux.server import Server
from tmuxp import config_reader

from .. import config, exc, log, util
from ..workspacebuilder import WorkspaceBuilder
from .utils import ConfigPath, _validate_choices, get_config_dir, tmuxp_echo


def set_layout_hook(session, hook_name):
    """Set layout hooks to normalize layout.

    References:

        - tmuxp issue: https://github.com/tmux-python/tmuxp/issues/309
        - tmux issue: https://github.com/tmux/tmux/issues/1106

    tmux 2.6+ requires that the window be viewed with the client before
    select-layout adjustments can take effect.

    To handle this, this function creates temporary hook for this session to
    iterate through all windows and select the layout.

    In order for layout changes to take effect, a client must at the very
    least be attached to the window (not just the session).

    hook_name is provided to allow this to set to multiple scenarios, such
    as 'client-attached' (which the user attaches the session). You may
    also want 'after-switch-client' for cases where the user loads tmuxp
    sessions inside tmux since tmuxp offers to switch for them.

    Also, the hooks are set immediately unbind after they're invoked via -u.

    Parameters
    ----------
    session : :class:`libtmux.session.Session`
        session to bind hook to
    hook_name : str
        hook name to bind to, e.g. 'client-attached'
    """
    cmd = ["set-hook", "-t", session.id, hook_name]
    hook_cmd = []
    attached_window = session.attached_window
    for window in session.windows:
        # unfortunately, select-layout won't work unless
        # we've literally selected the window at least once
        # with the client
        hook_cmd.append(f"selectw -t {window.id}")
        # edit: removed -t, or else it won't respect main-pane-w/h
        hook_cmd.append("selectl")
        hook_cmd.append("selectw -p")

    # unset the hook immediately after executing
    hook_cmd.append(
        "set-hook -u -t {target_session} {hook_name}".format(
            target_session=session.id, hook_name=hook_name
        )
    )
    hook_cmd.append(f"selectw -t {attached_window.id}")

    # join the hook's commands with semicolons
    hook_cmd = "{}".format("; ".join(hook_cmd))

    # append the hook command
    cmd.append(hook_cmd)

    # create the hook
    session.cmd(*cmd)


def load_plugins(sconf):
    """
    Load and return plugins in config
    """
    plugins = []
    if "plugins" in sconf:
        for plugin in sconf["plugins"]:
            try:
                module_name = plugin.split(".")
                module_name = ".".join(module_name[:-1])
                plugin_name = plugin.split(".")[-1]
                plugin = getattr(importlib.import_module(module_name), plugin_name)
                plugins.append(plugin())
            except exc.TmuxpPluginException as error:
                if not click.confirm(
                    "%sSkip loading %s?"
                    % (click.style(str(error), fg="yellow"), plugin_name),
                    default=True,
                ):
                    click.echo(
                        click.style("[Not Skipping] ", fg="yellow")
                        + "Plugin versions constraint not met. Exiting..."
                    )
                    sys.exit(1)
            except Exception as error:
                click.echo(
                    click.style("[Plugin Error] ", fg="red")
                    + f"Couldn't load {plugin}\n"
                    + click.style(f"{error}", fg="yellow")
                )
                sys.exit(1)

    return plugins


def _reattach(builder):
    """
    Reattach session (depending on env being inside tmux already or not)

    Parameters
    ----------
    builder: :class:`workspacebuilder.WorkspaceBuilder`

    Notes
    -----
    If ``TMUX`` environmental variable exists in the environment this script is
    running, that means we're in a tmux client. So ``tmux switch-client`` will
    load the session.

    If not, ``tmux attach-session`` loads the client to the target session.
    """
    for plugin in builder.plugins:
        plugin.reattach(builder.session)
        proc = builder.session.cmd("display-message", "-p", "'#S'")
        for line in proc.stdout:
            print(line)

    if "TMUX" in os.environ:
        builder.session.switch_client()

    else:
        builder.session.attach_session()


def _load_attached(builder, detached):
    """
    Load config in new session

    Parameters
    ----------
    builder: :class:`workspacebuilder.WorkspaceBuilder`
    detached : bool
    """
    builder.build()

    if "TMUX" in os.environ:  # tmuxp ran from inside tmux
        # unset TMUX, save it, e.g. '/tmp/tmux-1000/default,30668,0'
        tmux_env = os.environ.pop("TMUX")

        if has_gte_version("2.6"):
            set_layout_hook(builder.session, "client-session-changed")

        builder.session.switch_client()  # switch client to new session

        os.environ["TMUX"] = tmux_env  # set TMUX back again
    else:
        if has_gte_version("2.6"):
            # if attaching for first time
            set_layout_hook(builder.session, "client-attached")

            # for cases where user switches client for first time
            set_layout_hook(builder.session, "client-session-changed")

        if not detached:
            builder.session.attach_session()


def _load_detached(builder):
    """
    Load config in new session but don't attach

    Parameters
    ----------
    builder: :class:`workspacebuilder.WorkspaceBuilder`
    """
    builder.build()

    if has_gte_version("2.6"):  # prepare for both cases
        set_layout_hook(builder.session, "client-attached")
        set_layout_hook(builder.session, "client-session-changed")

    print("Session created in detached state.")


def _load_append_windows_to_current_session(builder):
    """
    Load config as new windows in current session

    Parameters
    ----------
    builder: :class:`workspacebuilder.WorkspaceBuilder`
    """
    current_attached_session = builder.find_current_attached_session()
    builder.build(current_attached_session, append=True)
    if has_gte_version("2.6"):  # prepare for both cases
        set_layout_hook(builder.session, "client-attached")
        set_layout_hook(builder.session, "client-session-changed")


def _setup_plugins(builder):
    """
    Runs after before_script

    Parameters
    ----------
    builder: :class:`workspacebuilder.WorkspaceBuilder`
    """
    for plugin in builder.plugins:
        plugin.before_script(builder.session)

    return builder.session


def load_workspace(
    config_file,
    socket_name=None,
    socket_path=None,
    tmux_config_file=None,
    new_session_name=None,
    colors=None,
    detached=False,
    answer_yes=False,
    append=False,
):
    """
    Load a tmux "workspace" session via tmuxp file.

    Parameters
    ----------
    config_file : str
        absolute path to config file
    socket_name : str, optional
        ``tmux -L <socket-name>``
    socket_path: str, optional
        ``tmux -S <socket-path>``
    new_session_name: str, options
        ``tmux new -s <new_session_name>``
    colors : str, optional
        '-2'
            Force tmux to support 256 colors
    detached : bool
        Force detached state. default False.
    answer_yes : bool
        Assume yes when given prompt to attach in new session.
        Default False.
    append : bool
       Assume current when given prompt to append windows in same session.
       Default False.

    Notes
    -----

    tmuxp will check and load a configuration file. The file will use ConfigReader
    to load a JSON/YAML into a :py:obj:`dict`. Then :func:`config.expand` and
    :func:`config.trickle` will be used to expand any shorthands, template
    variables, or file paths relative to where the config/script is executed
    from.

    :func:`config.expand` accepts the directory of the config file, so the
    user's configuration can resolve absolute paths relative to where the
    config file is. In otherwords, if a config file at */var/moo/hi.yaml*
    has *./* in its configs, we want to be sure any file path with *./* is
    relative to */var/moo*, not the user's PWD.

    A :class:`libtmux.Server` object is created. No tmux server is started yet,
    just the object.

    The prepared configuration and the server object is passed into an instance
    of :class:`~tmuxp.workspacebuilder.WorkspaceBuilder`.

    A sanity check against :meth:`libtmux.common.which` is ran. It will raise
    an exception if tmux isn't found.

    If a tmux session under the same name as ``session_name`` in the tmuxp
    configuration exists, tmuxp offers to attach the session. Currently, tmuxp
    does not allow appending a workspace / incremental building on top of a
    current session (pull requests are welcome!).

    :meth:`~tmuxp.workspacebuilder.WorkspaceBuilder.build` will build the session in
    the background via using tmux's detached state (``-d``).

    After the session (workspace) is built, unless the user decided to load
    the session in the background via ``tmuxp -d`` (which is in the spirit
    of tmux's ``-d``), we need to prompt the user to attach the session.

    If the user is already inside a tmux client, which we detect via the
    ``TMUX`` environment variable bring present, we will prompt the user to
    switch their current client to it.

    If they're outside of tmux client - in a plain-old PTY - we will
    automatically ``attach``.

    If an exception is raised during the building of the workspace, tmuxp will
    prompt to cleanup (``$ tmux kill-session``) the session on the user's
    behalf. An exception raised during this process means it's not easy to
    predict how broken the session is.

    .. versionchanged:: tmux 2.6+

        In tmux 2.6, the way layout and proportion's work when interfacing
        with tmux in a detached state (outside of a client) changed. Since
        tmuxp builds workspaces in a detached state, the WorkspaceBuilder isn't
        able to rely on functionality requiring awarness of session geometry,
        e.g. ``set-layout``.

        Thankfully, tmux is able to defer commands to run after the user
        performs certain actions, such as loading a client via
        ``attach-session`` or ``switch-client``.

        Upon client switch, ``client-session-changed`` is triggered [1]_.

    References
    ----------
    .. [1] cmd-switch-client.c hook. GitHub repo for tmux.

       https://github.com/tmux/tmux/blob/2.6/cmd-switch-client.c#L132.
       Accessed April 8th, 2018.
    """
    # get the canonical path, eliminating any symlinks
    if isinstance(config_file, str):
        config_file = pathlib.Path(config_file)

    tmuxp_echo(
        click.style("[Loading] ", fg="green")
        + click.style(str(config_file), fg="blue", bold=True)
    )

    # ConfigReader allows us to open a yaml or json file as a dict
    raw_config = config_reader.ConfigReader._from_file(config_file)
    # shapes configurations relative to config / profile file location
    sconfig = config.expand(raw_config, cwd=os.path.dirname(config_file))
    # Overwrite session name
    if new_session_name:
        sconfig["session_name"] = new_session_name
    # propagate config inheritance (e.g. session -> window, window -> pane)
    sconfig = config.trickle(sconfig)

    t = Server(  # create tmux server object
        socket_name=socket_name,
        socket_path=socket_path,
        config_file=tmux_config_file,
        colors=colors,
    )

    shutil.which("tmux")  # raise exception if tmux not found

    try:  # load WorkspaceBuilder object for tmuxp config / tmux server
        builder = WorkspaceBuilder(
            sconf=sconfig, plugins=load_plugins(sconfig), server=t
        )
    except exc.EmptyConfigException:
        tmuxp_echo("%s is empty or parsed no config data" % config_file, err=True)
        return

    session_name = sconfig["session_name"]

    # if the session already exists, prompt the user to attach
    if builder.session_exists(session_name) and not append:
        if not detached and (
            answer_yes
            or click.confirm(
                "%s is already running. Attach?"
                % click.style(session_name, fg="green"),
                default=True,
            )
        ):
            _reattach(builder)
        return

    try:
        if detached:
            _load_detached(builder)
            return _setup_plugins(builder)

        if append:
            if "TMUX" in os.environ:  # tmuxp ran from inside tmux
                _load_append_windows_to_current_session(builder)
            else:
                _load_attached(builder, detached)

            return _setup_plugins(builder)

        # append and answer_yes have no meaning if specified together
        elif answer_yes:
            _load_attached(builder, detached)
            return _setup_plugins(builder)

        if "TMUX" in os.environ:  # tmuxp ran from inside tmux
            msg = (
                "Already inside TMUX, switch to session? yes/no\n"
                "Or (a)ppend windows in the current active session?\n[y/n/a]"
            )
            options = ["y", "n", "a"]
            choice = click.prompt(msg, value_proc=_validate_choices(options))

            if choice == "y":
                _load_attached(builder, detached)
            elif choice == "a":
                _load_append_windows_to_current_session(builder)
            else:
                _load_detached(builder)
        else:
            _load_attached(builder, detached)

    except exc.TmuxpException as e:
        import traceback

        tmuxp_echo(traceback.format_exc(), err=True)
        tmuxp_echo(e, err=True)

        choice = click.prompt(
            "Error loading workspace. (k)ill, (a)ttach, (d)etach?",
            value_proc=_validate_choices(["k", "a", "d"]),
            default="k",
        )

        if choice == "k":
            builder.session.kill_session()
            tmuxp_echo("Session killed.")
        elif choice == "a":
            _reattach(builder)
        else:
            sys.exit()

    return _setup_plugins(builder)


def config_file_completion(ctx, params, incomplete):
    config_dir = pathlib.Path(get_config_dir())
    choices: List[pathlib.Path] = []

    # CWD Paths
    choices += sorted(
        pathlib.Path(os.path.relpath(p, pathlib.Path.cwd()))
        for p in [pathlib.Path.cwd(), *pathlib.Path.cwd().parents]
        if config.in_dir(str(p)) or len(list(p.glob(".tmuxp.*")))
    )
    # CWD look one directory up
    choices += [
        pathlib.Path(f"./{os.path.relpath(p, pathlib.Path.cwd())}")
        for p in pathlib.Path.cwd().glob("*/.tmuxp.*")
    ]

    # Project configs
    choices += sorted((config_dir / c).stem for c in config.in_dir(str(config_dir)))

    return sorted(str(c) for c in choices if str(c).startswith(incomplete))


@click.command(name="load", short_help="Load tmuxp workspaces.")
@click.pass_context
@click.argument(
    "config",
    type=ConfigPath(exists=True),
    nargs=-1,
    shell_complete=config_file_completion,
)
@click.option("-S", "socket_path", help="pass-through for tmux -S")
@click.option("-L", "socket_name", help="pass-through for tmux -L")
@click.option("-f", "tmux_config_file", help="pass-through for tmux -f")
@click.option("-s", "new_session_name", help="start new session with new session name")
@click.option("--yes", "-y", "answer_yes", help="yes", is_flag=True)
@click.option(
    "-d", "detached", help="Load the session without attaching it", is_flag=True
)
@click.option(
    "-a",
    "append",
    help="Load configuration, appending windows to the current session",
    is_flag=True,
)
@click.option(
    "colors",
    "-2",
    flag_value=256,
    default=True,
    help="Force tmux to assume the terminal supports 256 colours.",
)
@click.option(
    "colors",
    "-8",
    flag_value=88,
    help="Like -2, but indicates that the terminal supports 88 colours.",
)
@click.option("--log-file", help="File to log errors/output to")
def command_load(
    ctx,
    config,
    socket_name,
    socket_path,
    tmux_config_file,
    new_session_name,
    answer_yes,
    detached,
    append,
    colors,
    log_file,
):
    """Load a tmux workspace from each CONFIG.

    CONFIG is a specifier for a configuration file.

    If CONFIG is a path to a directory, tmuxp will search it for
    ".tmuxp.{yaml,yml,json}".

    If CONFIG is has no directory component and only a filename, e.g.
    "myconfig.yaml", tmuxp will search the users's config directory for that
    file.

    If CONFIG has no directory component, and only a name with no extension,
    e.g. "myconfig", tmuxp will search the users's config directory for any
    file with the extension ".yaml", ".yml", or ".json" that matches that name.

    If multiple configuration files that match a given CONFIG are found, tmuxp
    will warn and pick the first one found.

    If multiple CONFIGs are provided, workspaces will be created for all of
    them. The last one provided will be attached. The others will be created in
    detached mode.
    """
    util.oh_my_zsh_auto_title()

    if log_file:
        logfile_handler = logging.FileHandler(log_file)
        logfile_handler.setFormatter(log.LogFormatter())
        from . import logger

        logger.addHandler(logfile_handler)

    tmux_options = {
        "socket_name": socket_name,
        "socket_path": socket_path,
        "tmux_config_file": tmux_config_file,
        "new_session_name": new_session_name,
        "answer_yes": answer_yes,
        "colors": colors,
        "detached": detached,
        "append": append,
    }

    if not config:
        tmuxp_echo("Enter at least one CONFIG")
        tmuxp_echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    if isinstance(config, str):
        load_workspace(config, **tmux_options)

    elif isinstance(config, tuple):
        config = list(config)
        # Load each configuration but the last to the background
        for cfg in config[:-1]:
            opt = tmux_options.copy()
            opt.update({"detached": True, "new_session_name": None})
            load_workspace(cfg, **opt)

        # todo: obey the -d in the cli args only if user specifies
        load_workspace(config[-1], **tmux_options)



"""Tests for .'s helper and utility functions."""
import pytest

from libtmux.test import get_test_session_name, temp_session


def test_kills_session(server):
    server = server
    session_name = get_test_session_name(server=server)

    with temp_session(server=server, session_name=session_name):
        result = server.has_session(session_name)
        assert result

    assert not server.has_session(session_name)


@pytest.mark.flaky(reruns=5)
def test_if_session_killed_before(server):
    """Handles situation where session already closed within context"""

    server = server
    session_name = get_test_session_name(server=server)

    with temp_session(server=server, session_name=session_name):

        # an error or an exception within a temp_session kills the session
        server.kill_session(session_name)

        result = server.has_session(session_name)
        assert not result

    # really dead?
    assert not server.has_session(session_name)
